import numpy as np
import pandas as pd
from pandas.util.testing import assert_frame_equal
import pytest
from scipy.spatial.distance import euclidean
from sklearn.preprocessing import MaxAbsScaler

from api.client.samples.analogous_years.lib import distance_matrix


# TODO: test_dtw_dist_matrix

def create_test_data():
    return pd.DataFrame({'rows': ['2019-04-01 to 2019-08-01',
                                  '2018-04-01 to 2018-08-01',
                                  '2017-04-01 to 2017-08-01'],
                         '2019-04-01 to 2019-08-01': [0, 40, 9],
                         '2018-04-01 to 2018-08-01': [40, 0, 44],
                         '2017-04-01 to 2017-08-01': [9, 44, 0]}). \
        set_index('rows')


def create_test_dates():
    return [{'initial_date': '2018-04-01', 'final_date': '2018-08-01'},
            {'initial_date': '2016-01-04', 'final_date': '2017-01-08'}]


def create_test_methods():
    return ['euclidean', 'invalid_name']


def test_euclidean_distance_matrix():
    test_data_1 = pd.DataFrame({'rows': ['row_1', 'row_2', 'row_3'],
                                'col_1': [1, 2, 3],
                                'col_2': [4, 5, 6],
                                'col_3': [7, 8, 9]}).set_index('rows')
    expected = np.zeros((test_data_1.shape[0], test_data_1.shape[0]))
    for i in range(test_data_1.shape[0]):
        for j in range(i, test_data_1.shape[0]):
            expected[i, j] = euclidean(test_data_1.iloc[i], test_data_1.iloc[j])
            expected[j, i] = euclidean(test_data_1.iloc[i], test_data_1.iloc[j])
    expected = pd.DataFrame(expected)
    expected.columns = test_data_1.index
    expected.index = test_data_1.index
    assert_frame_equal(distance_matrix.euclidean_dist_matrix(test_data_1), expected)


def test_scaled_labeled_method_distances():
    initial_date = create_test_dates()[0]['initial_date']
    final_date = create_test_dates()[0]['final_date']
    method = create_test_methods()[0]
    admissible_test_date = initial_date + ' to ' + final_date
    expected = create_test_data()
    expected = expected[[admissible_test_date]]
    expected.rename(columns={admissible_test_date:method}, inplace=True)
    scaler = MaxAbsScaler()
    expected.loc[:, method] = scaler.fit_transform(expected[[method]])
    assert_frame_equal(distance_matrix.scaled_labeled_method_distances
                       (create_test_data(), initial_date, final_date,
                        method), expected)


def test_scaled_labeled_method_distances_invalid_date():
    invalid_initial_date = create_test_dates()[1]['initial_date']
    invalid_final_date = create_test_dates()[1]['final_date']
    method = create_test_methods()[0]
    with pytest.raises(KeyError):
        distance_matrix.scaled_labeled_method_distances(create_test_data(),
                                                        invalid_initial_date,
                                                        invalid_final_date,
                                                        method)


def test_scaled_labeled_method_distances_invalid_method():
    initial_date = create_test_dates()[0]['initial_date']
    final_date = create_test_dates()[0]['final_date']
    invalid_method = create_test_methods()[1]
    with pytest.raises(ValueError):
        distance_matrix.scaled_labeled_method_distances(create_test_data(),
                                                        initial_date,
                                                        final_date,
                                                        invalid_method)



try:
    # Python 3.3+
    from unittest.mock import patch, MagicMock
except ImportError:
    # Python 2.7
    from mock import patch, MagicMock
from unittest import TestCase
from datetime import date

from groclient import GroClient
from groclient.mock_data import mock_entities, mock_data_series, mock_data_points

MOCK_HOST = "pytest.groclient.url"
MOCK_TOKEN = "pytest.groclient.token"


def mock_get_available(access_token, api_host, entity_type):
    return list(mock_entities[entity_type].values())


def mock_list_available(access_token, api_host, selected_entities):
    return [dict(data_series) for data_series in mock_data_series]


def mock_lookup(access_token, api_host, entity_type, entity_ids):
    if isinstance(entity_ids, int):
        # Raises a KeyError if the requested entity hasn't been mocked:
        return mock_entities[entity_type][entity_ids]
    else:
        return {
            str(entity_id): mock_entities[entity_type][entity_id]
            for entity_id in entity_ids
        }


def mock_get_allowed_units(access_token, api_host, metric_id, item_id):
    return [unit["id"] for unit in mock_entities["units"].values()]


def mock_get_data_series(access_token, api_host, **selection):
    return [dict(data_series) for data_series in mock_data_series]


def mock_search(access_token, api_host, entity_type, search_terms):
    return [
        {"id": entity["id"]}
        for entity in mock_entities[entity_type].values()
        if search_terms in entity["name"]
    ]


def mock_rank_series_by_source(access_token, api_host, selections_list):
    for data_series in mock_data_series:
        yield data_series


def mock_get_geo_centre(access_token, api_host, region_id):
    return [
         {"centre": [45.7228, -112.996], "regionId": 1215, "regionName": "United States"}
    ]


def mock_get_geojson(access_token, api_host, region_id, zoom_level):
    if zoom_level < 7:
        return {
            "type": "GeometryCollection",
            "geometries": [
                {"type": "MultiPolygon", "coordinates": [[[[-38, -4]]]]}
            ],
        }
    else:
        return {
            "type": "GeometryCollection",
            "geometries": [
                {"type": "MultiPolygon", "coordinates": [[[[-38.394, -4.225]]]]}
            ],
        }


def mock_get_geojsons(access_token, api_host, region_id, descendant_level, zoom_level):
    return [{'region_id': 13051, 'region_name': 'Alabama', 'centre': [32.7933, -86.8278],
             'geojson': {'type': 'MultiPolygon',
                         'coordinates': [[[[-88.201896667, 35.0088806150001], [-88.079490661, 35.006961823], [-87.987052917, 35.0075187690001], [-87.610366821, 35.0048713690001]]]]}},
            {'region_id': 13052, 'region_name': 'Alaska', 'centre': [64.2386, -152.279],
             'geojson': {'type': 'MultiPolygon',
                         'coordinates': [[[[-179.07043457, 51.2564086920001], [-179.094436645, 51.2270851140001], [-179.142776489, 51.2288894660001]]]]}}
    ]


def mock_get_descendant(
    access_token,
    api_host,
    entity_type,
    entity_id,
    distance,
    include_details,
):

    childs = [
            child
            for child in mock_entities[entity_type].values()
            if 119 in child["belongsTo"]
        ]
    if include_details:
        return childs
    else:
        return [{"id": child["id"]} for child in childs]


def mock_get_descendant_regions(
    access_token,
    api_host,
    region_id,
    descendant_level,
    include_historical,
    include_details,
):
    if descendant_level:
        regions = [
            region
            for region in mock_entities["regions"].values()
            if region["level"] == descendant_level
        ]
    else:
        regions = list(mock_entities["regions"].values())
    if not include_historical:
        regions = [region for region in regions if not region["historical"]]
    if include_details:
        return regions
    else:
        return [{"id": region["id"]} for region in regions]


def mock_get_available_timefrequency(access_token, api_host, **selection):
    return [
        {
            "start_date": "2000-02-18T00:00:00.000Z",
            "frequency_id": 3,
            "end_date": "2020-03-12T00:00:00.000Z",
            "name": "8-day",
        },
        {
            "start_date": "2019-09-02T00:00:00.000Z",
            "frequency_id": 1,
            "end_date": "2020-03-09T00:00:00.000Z",
            "name": "daily",
        },
    ]


def mock_get_top(access_token, api_host, entity_type, num_results, **selection):
    return [
        {
            "metricId": 860032,
            "itemId": 274,
            "regionId": 1215,
            "frequencyId": 9,
            "sourceId": 2,
            "value": 400,
            "unitId": 14,
        },
        {
            "metricId": 860032,
            "itemId": 274,
            "regionId": 1215,
            "frequencyId": 9,
            "sourceId": 2,
            "value": 395,
            "unitId": 14,
        },
        {
            "metricId": 860032,
            "itemId": 274,
            "regionId": 1215,
            "frequencyId": 9,
            "sourceId": 2,
            "value": 12,
            "unitId": 14,
        },
    ]


def mock_get_data_points(access_token, api_host, **selections):
    if isinstance(selections["region_id"], int):
        data_point = dict(mock_data_points[0])
        # set the data_point to use the selected region
        # other ids in mocked data points may not line up with selected ids
        data_point["region_id"] = selections["region_id"]
        return [data_point]
    elif isinstance(selections["region_id"], list):
        data_points = []
        for idx, region_id in enumerate(selections["region_id"]):
            # use mock_data_points[0] or mock_data_points[1] as a base depending if
            # index is even or odd
            data_point = dict(mock_data_points[idx % 2])
            data_point["region_id"] = region_id
            data_points.append(data_point)
        return data_points


@patch("groclient.lib.get_available", MagicMock(side_effect=mock_get_available))
@patch("groclient.lib.list_available", MagicMock(side_effect=mock_list_available))
@patch("groclient.lib.lookup", MagicMock(side_effect=mock_lookup))
@patch(
    "groclient.lib.get_allowed_units", MagicMock(side_effect=mock_get_allowed_units)
)
@patch("groclient.lib.get_data_series", MagicMock(side_effect=mock_get_data_series))
@patch("groclient.lib.search", MagicMock(side_effect=mock_search))
@patch(
    "groclient.lib.rank_series_by_source",
    MagicMock(side_effect=mock_rank_series_by_source),
)
@patch("groclient.lib.get_geo_centre", MagicMock(side_effect=mock_get_geo_centre))
@patch("groclient.lib.get_geojsons", MagicMock(side_effect=mock_get_geojsons))
@patch("groclient.lib.get_geojson", MagicMock(side_effect=mock_get_geojson))
@patch("groclient.lib.get_descendant", MagicMock(side_effect=mock_get_descendant))
@patch(
    "groclient.lib.get_descendant_regions",
    MagicMock(side_effect=mock_get_descendant_regions),
)
@patch(
    "groclient.lib.get_available_timefrequency",
    MagicMock(side_effect=mock_get_available_timefrequency),
)
@patch("groclient.lib.get_top", MagicMock(side_effect=mock_get_top))
@patch("groclient.lib.get_data_points", MagicMock(side_effect=mock_get_data_points))
class GroClientTests(TestCase):
    def setUp(self):
        self.client = GroClient(MOCK_HOST, MOCK_TOKEN)
        self.client._async_http_client = None  # Force tests to use synchronous http
        self.assertTrue(isinstance(self.client, GroClient))

    def test_get_logger(self):
        # should NOT raise any exception if get_logger correctly returns a logger object:
        self.client.get_logger().debug("Test output")

    def test_get_available(self):
        self.assertTrue("name" in self.client.get_available("units")[0])

    def test_list_available(self):
        self.assertTrue("metric_id" in self.client.list_available({})[0])

    def test_lookup(self):
        self.assertEqual(self.client.lookup("units", 10)["name"], "kilogram")

    def test_lookup_unit_abbreviation(self):
        self.assertEqual(self.client.lookup_unit_abbreviation(10), "kg")

    def test_get_allowed_units(self):
        self.assertTrue(isinstance(self.client.get_allowed_units(1, 1)[0], int))

    def test_get_data_series(self):
        self.assertTrue("metric_id" in self.client.get_data_series()[0])

    def test_search(self):
        self.assertTrue(isinstance(self.client.search("regions", "United")[0], dict))
        self.assertTrue("id" in self.client.search("regions", "United")[0])

    def test_search_and_lookup(self):
        region = next(self.client.search_and_lookup("regions", "United"))
        self.assertTrue(isinstance(region, dict))
        self.assertTrue("id" in region)
        self.assertEqual(region["name"], "United States")

    def test_lookup_belongs(self):
        self.assertEqual(
            next(self.client.lookup_belongs("regions", 1215))["name"], "World"
        )

    def test_rank_series_by_source(self):
        series = next(self.client.rank_series_by_source([]))
        self.assertTrue("metric_id" in series)
        self.assertTrue("source_id" in series)

    def test_get_geo_centre(self):
        centre = self.client.get_geo_centre(1215)
        self.assertTrue(len(centre) == 1)
        self.assertTrue("centre" in centre[0])
        self.assertTrue("regionName" in centre[0])

    def test_get_geojsons(self):
        geojsons = self.client.get_geojsons(1215, 4)
        self.assertTrue(len(geojsons) == 2)
        self.assertTrue("region_id" in geojsons[0])
        self.assertTrue("region_name" in geojsons[0])
        self.assertTrue("centre" in geojsons[0])
        self.assertTrue("geojson" in geojsons[0])
        self.assertTrue("type" in geojsons[0]["geojson"])
        self.assertTrue("coordinates" in geojsons[0]["geojson"])

    def test_get_geojson(self):
        geojson = self.client.get_geojson(1215)
        self.assertTrue("type" in geojson)
        self.assertTrue("type" in geojson["geometries"][0])
        self.assertTrue("coordinates" in geojson["geometries"][0])
        self.assertTrue(geojson["geometries"][0]['coordinates'][0][0][0] == [-38.394, -4.225])
        geojson = self.client.get_geojson(1215, 1)
        self.assertTrue("type" in geojson)
        self.assertTrue("type" in geojson["geometries"][0])
        self.assertTrue("coordinates" in geojson["geometries"][0])
        self.assertTrue(geojson["geometries"][0]['coordinates'][0][0][0] == [-38, -4])

    def test_get_descendant(self):
        self.assertTrue("name" in self.client.get_descendant('metrics', 119)[0])
        self.assertTrue(
            "name"
            not in self.client.get_descendant('metrics', 119, include_details=False)[0]
        )

    def test_get_descendant_regions(self):
        self.assertTrue("name" in self.client.get_descendant_regions(1215)[0])
        self.assertTrue(
            "name"
            not in self.client.get_descendant_regions(1215, include_details=False)[0]
        )

    def test_get_available_timefrequency(self):
        self.assertEqual(
            self.client.get_available_timefrequency()[0]["frequency_id"], 3
        )

    def test_get_top(self):
        self.assertEqual(
            self.client.get_top(
                "regions", metric_id=860032, item_id=274, frequency_id=9, source_id=2
            )[0]["regionId"],
            1215,
        )

    def test_get_df(self):
        self.client.add_single_data_series(mock_data_series[0])
        df = self.client.get_df()
        self.assertEqual(df.iloc[0]["start_date"].date(), date(2017, 1, 1))
        self.client.add_single_data_series(mock_data_series[0])
        df = self.client.get_df(show_revisions=True)
        self.assertEqual(df.iloc[0]["start_date"].date(), date(2017, 1, 1))
        indexed_df = self.client.get_df(index_by_series=True)
        self.assertEqual(indexed_df.iloc[0]["start_date"].date(), date(2017, 1, 1))
        series = dict(zip(indexed_df.index.names, indexed_df.iloc[0].name))
        self.assertEqual(series, mock_data_series[0])

    def test_get_df_show_revisions(self):
        self.client.add_single_data_series(mock_data_series[0])
        df = self.client.get_df(show_revisions=True)
        self.assertEqual(df.iloc[0]["start_date"].date(), date(2017, 1, 1))

    def test_add_points_to_df(self):
        self.client.add_points_to_df(None, mock_data_series[0], [])
        self.assertTrue(self.client.get_df().empty)
        self.assertTrue(self.client.get_df(show_revisions=True).empty)
        self.assertTrue(self.client.get_df(index_by_series=True).empty)

        data_points = self.client.get_data_points(**mock_data_series[0])
        self.client.add_points_to_df(None, mock_data_series[0], data_points)
        self.assertEqual(
            self.client.get_df().iloc[0]["start_date"].date(), date(2017, 1, 1)
        )
        self.assertEqual(
            self.client.get_df().iloc[0]["source_id"], 2
        )

    def test_get_data_points(self):
        # Gives the point's default unit if unit's not specified:
        data_points = self.client.get_data_points(**mock_data_series[0])
        self.assertEqual(data_points[0]["unit_id"], 14)
        self.assertEqual(data_points[0]["value"], 40891)

        # Converts to the given unit:
        selections = dict(
            mock_data_series[0]
        )  # make a copy so we don't modify the original
        selections["unit_id"] = 10
        data_points = self.client.get_data_points(**selections)
        self.assertEqual(data_points[0]["unit_id"], 10)
        self.assertEqual(data_points[0]["value"], 40891000)

    def test_GDH(self):
        df = self.client.GDH("860032-274-1215-0-9-2")
        self.assertEqual(df.iloc[0]["start_date"].date(), date(2017, 1, 1))
        # if you request a series with no data, an empty dataframe should be returned:
        # Extra options can be given, but value in the GDH key itself (metric_id/item_id/etc.)
        # should be ignored.
        df = self.client.GDH("860032-274-1215-0-2-9", insert_nulls=True, metric_id=1)
        self.assertEqual(len(df), 0)

    def test_add_single_data_series_adds_copy(self):
        selections = dict(mock_data_series[0])  # don't modify test data. Make a copy
        for region_id in [
            mock_data_series[0]["region_id"],
            mock_data_series[1]["region_id"],
        ]:
            # modify the original selections object
            selections["region_id"] = region_id
            # if add_single_data_series isn't making a copy of the selections passed in,
            # then this test should fail since the original reference has been modified.
            self.client.add_single_data_series(selections)
        self.assertEqual(
            len(self.client.get_df().drop_duplicates().region_id.unique()), 2
        )

    def test_add_single_data_series_allows_metadata(self):
        selections = dict(mock_data_series[0])
        selections['metadata'] = {'includes_historical_region': True}
        self.client.add_single_data_series(selections)
        self.assertEqual(len(self.client.get_df().item_id), 1)

    def test_get_data_series_list(self):
        self.client.add_single_data_series(mock_data_series[0])
        for key, value in self.client.get_data_series_list()[0].items():
            self.assertEqual(value, mock_data_series[0][key])

    def test_find_data_series(self):
        # TODO: when duplicates are removed, this should equal 2:
        self.assertEqual(
            len(
                list(
                    self.client.find_data_series(
                        metric="Production",
                        region="United",
                        start_date="2000-01-01",
                        end_date="2005-12-31",
                    )
                )
            ),
            8,
        )

        # TODO: when duplicates are removed, this should equal 2:
        def only_accept_production_quantity(search_result):
            return "metric_id" not in search_result or search_result["metric_id"] == 860032
        self.assertEqual(
            len(
                list(
                    self.client.find_data_series(
                        metric="Production",
                        result_filter=only_accept_production_quantity
                    )
                )
            ),
            8,
        )

    def test_add_data_series(self):
        # TODO: when duplicates are removed, this should equal 2:
        data_series = self.client.add_data_series(metric="Production", region="United")
        self.assertEqual(data_series, mock_data_series[0])
        for key, value in self.client.get_data_series_list()[0].items():
            self.assertEqual(value, mock_data_series[0][key])

    def test_search_for_entity(self):
        self.assertEqual(self.client.search_for_entity("metrics", "Production"), 860032)

    def test_get_provinces(self):
        self.assertEqual(
            self.client.get_provinces("United")[0], mock_entities["regions"][12345]
        )

    def test_get_names_for_selection(self):
        selection = {"metric_id": 860032, "region_id": 0}
        self.assertEqual(
            self.client.get_names_for_selection(selection),
            [("metric", "Production Quantity"), ("region", "World")],
        )

    def test_convert_unit(self):
        self.assertEqual(
            self.client.convert_unit({"value": 1, "unit_id": 10}, 10),
            {"value": 1, "unit_id": 10},
        )
        self.assertEqual(
            self.client.convert_unit({"value": 1, "unit_id": 10}, 14),
            {"value": 0.001, "unit_id": 14},
        )
        self.assertEqual(
            self.client.convert_unit({"value": 3, "unit_id": 36}, 37),
            {"value": 42, "unit_id": 37},
        )
        self.assertEqual(
            self.client.convert_unit({"value": 1, "unit_id": 37}, 36),
            {"value": -17.5, "unit_id": 36},
        )
        self.assertEqual(
            self.client.convert_unit({"value": 20, "unit_id": 10, "metadata": {"conf_interval": 2}}, 14),
            {"value": 0.02, "metadata": {"conf_interval": 0.002}, "unit_id": 14},
        )
        self.assertEqual(
            self.client.convert_unit({"value": 20, "unit_id": 10, "metadata": {}}, 14),
            {"value": 0.02, "metadata": {}, "unit_id": 14},
        )

        self.assertEqual(self.client.convert_unit({}, 36), {})

        with self.assertRaises(Exception):
            self.client.convert_unit({"value": 1, "unit_id": 10}, 43)

        self.assertEqual(
            self.client.convert_unit({"value": None, "unit_id": 37}, 36),
            {"value": None, "unit_id": 36},
        )

        with self.assertRaises(Exception):
            self.client.convert_unit({"value": None, "unit_id": 10}, 43)



def top_k(dataframe, column_name, k = 5):
    time_period_list = []
    for i in range(2, 2+k):
        time_period_list.append(
            dataframe[dataframe[column_name] == i].index.tolist()[0])
    return set(time_period_list)



"""
This library contains utilities for
1. consolidating the final ranks
2. saving a dataframe in a chosen location
"""
from functools import reduce
import os

import matplotlib
matplotlib.use('agg')
import seaborn as sns

import numpy as np
import pandas as pd

from groclient import GroClient
from api.client.samples.analogous_years.lib import \
    distance_matrix, \
    feature_extractions, \
    get_transform_data

DEFAULT_API_HOST = 'api.gro-intelligence.com'


def common_start_date(client, data_series, provided_start_date=None):
    """Computes the earliest available start date from which all gro-data_series have data"""
    logger = client.get_logger()
    start_date_list = []
    if provided_start_date:
        start_date_list.append(provided_start_date)
    for i in range(len(data_series)):
        dates = client.get_data_points(**data_series[i])
        if len(dates) == 0:
            msg = "No data found for the following gro-data_series - {}".format(data_series[i])
            logger.warning(msg)
            raise Exception
        else:
            start_date_list.append(dates[0]['start_date'])
    start_date = max(start_date_list)
    for i in range(len(data_series)):
        data_series[i]['start_date'] = start_date
    return {'data_series': data_series, 'start_date': start_date}


def enso_data(start_date):
    enso_data_series = {'metric_id': 15851977,
                        'item_id': 13495,
                        'region_id': 0,
                        'source_id': 124,
                        'start_date': start_date,
                        'frequency_id': 6}
    return enso_data_series


def get_file_name(api_token, data_series_list, initial_date, final_date,
                  api_host=DEFAULT_API_HOST):
    """Combines region, items, and dates to return a string"""
    client = GroClient(api_host, api_token)
    logger = client.get_logger()
    key_words = [client.lookup('regions', data_series_list[0]['region_id'])['name']]
    for i in range(len(data_series_list)):
        key_words.append(client.lookup('items', data_series_list[i]['item_id'])['name'])
    key_words.append(initial_date)
    key_words.append(final_date)
    logger.info("\n Computing analogous years' ranks in {} with respect to {} between {} and {} \n".
                format(key_words[0], key_words[1:-2], initial_date, final_date))
    combined_name = '_'.join(key_words)
    return combined_name


def time_series(client, data_series, initial_date, final_date):
    """
    retrieves a sub-time series of a time series associated with a gro data_series
    :param client: GroClient
    :param data_series: A dictionary of Gro data_series
    :param final_date: 'YYYY-MM-DD'
    :param initial_date: 'YYYY-MM-DD'
    :return: A dataframe with data from the relevant dates
    """
    logger = client.get_logger()
    entities = {'metric_id', 'item_id', 'region_id', 'source_id', 'frequency_id', 'start_date'}
    data_series = {k: data_series[k] for k in data_series if k in entities}
    data = get_transform_data.get_data(client, **data_series)
    try:
        ts = get_transform_data.extract_time_periods_by_dates(data, initial_date, final_date)
        return ts
    except Exception as e:
        message = ('Please check availability of data for {}'.format(client.lookup(
            'items', data_series['item_id'])['name']))
        logger.warning(message)
        raise e


def ay_tsfresh(timeseries, num_jobs=0):
    tsfresh_features = feature_extractions.ts_feature_extraction(
        timeseries, num_jobs=num_jobs)
    return distance_matrix.euclidean_dist_matrix(tsfresh_features)


def ay_cumulative(timeseries, group='period', value='value'):
    cumulative_value = feature_extractions.cumulative(timeseries, group, value)
    return distance_matrix.euclidean_dist_matrix(cumulative_value)


def ay_euclidean(timeseries):
    transposed_pivot_time_series = get_transform_data.stack_time_periods_by_ddmm(
        timeseries).transpose()
    return distance_matrix.euclidean_dist_matrix(transposed_pivot_time_series)


def ay_dtw(timeseries):
    pivot_time_series = get_transform_data.stack_time_periods_by_ddmm(timeseries)
    return distance_matrix.dtw_dist_matrix(pivot_time_series)


def ranked_df_dictionary(client, data_series, initial_date, final_date, item, methods_list,
                         tsfresh_num_jobs=0):
    """
    Returns a dictionary whose keys are given by a 'method_item' and values are the dataframes
    of distances computed for that item using that method. Example: key: cumulative_Rainfall
    :param client: GroClient
    :param data_series: A dictionary of Gro data_series
    :param initial_date: 'YYYY-MM-DD'
    :param final_date: 'YYYY-MM-DD'
    :param item: Gro Item
    :param methods_list: a sublist of ['cumulative', 'euclidean', 'ts-features', 'dtw']
    :param tsfresh_num_jobs: integer, number of parallel processes in tsfresh
    :return: Dictionary of dataframes
    """
    ts = time_series(client, data_series, initial_date, final_date)
    dictionary_of_methods = {'cumulative_' + item: ay_cumulative,
                             'euclidean_' + item: ay_euclidean,
                             'dtw_' + item: ay_dtw,
                             'ts-features_' + item: ay_tsfresh}
    ranked_dfs = {}
    for key, value in dictionary_of_methods.items():
        if key.split('_')[0] in methods_list:
            if key.split('_')[0] == 'ts-features':
                ranked_dfs[key] = distance_matrix.scaled_labeled_method_distances(
                    value(ts, num_jobs=tsfresh_num_jobs), initial_date, final_date, key)
            else:
                ranked_dfs[key] = distance_matrix.scaled_labeled_method_distances(
                    value(ts), initial_date, final_date, key)
    return ranked_dfs


def merge_using_index_column(left, right):
    """
    Merges two dataframes using the indices
    :param left: pandas dataframe
    :param right: pandas dataframe
    :return: pandas dataframe
    """
    return pd.merge(left, right, left_index=True, right_index=True)


def combined_methods_distances(dictionary_of_df):
    """
    Merge all the df listed in the dictionary into a single dataframe containing the scaled
    distances. The keys of the dictionary in this case are the methods
    used to obtain the distances
    :param dictionary_of_df: A dictionary of pandas dataframes
    :return: A dataframe
    """
    rank_df = reduce(merge_using_index_column, list(dictionary_of_df.values()))
    return rank_df


def analogous_years(api_token, data_series_list, initial_date, final_date,
                    methods_list=['euclidean', 'cumulative', 'ts-features'],
                    all_ranks=None, weights=None, enso=None, enso_weight=None,
                    provided_start_date=None, tsfresh_num_jobs=0,
                    api_host=DEFAULT_API_HOST):
    """
    Use L^2 distance function to combine weighted distances from multiple gro-data_series
    and return the rank
    :param api_token: string, Gro-api token
    :param data_series_list: list of dictionaries containing gro data series
    :param initial_date: A date in YYYY-MM-DD format
    :param final_date: A date in YYYY-MM-DD format
    :param methods_list: a sublist of ['cumulative', 'euclidean', 'ts-features', 'dtw']
    :param all_ranks: Boolean to determine if all ranks will be displayed or a composite rank
    :param weights: Float determining the weight given to each data_series
    :param enso: Boolean to include ENSO
    :param enso_weight: Float
    :param provided_start_date: A string in YYYY-MM-DD format
    :param tsfresh_num_jobs: integer, number of parallel processes in tsfresh
    :param api_host:
    :return: A tuple (string, dataframe)
    The string contains '_' separated region, item, date
    The dataframe contains integer values (ranks)
    """
    client = GroClient(api_host, api_token)
    combined_items_distances = None
    data_series_list = common_start_date(client, data_series_list, provided_start_date)[
        'data_series']
    start_date = common_start_date(client, data_series_list, provided_start_date)[
        'start_date']
    if not weights:
        weights = [1] * len(data_series_list)
    if enso:
        data_series_list.append(enso_data(start_date))
        if enso_weight:
            weights.append(enso_weight)
        else:
            weights.append(1)
    for i in range(len(data_series_list)):
        gro_item = client.lookup('items', data_series_list[i]['item_id'])['name']
        combined_methods_distances_df = combined_methods_distances(
            ranked_df_dictionary(
                client, data_series_list[i], initial_date,
                final_date, gro_item, methods_list, tsfresh_num_jobs=tsfresh_num_jobs))
        numpy_combined_methods_distances = combined_methods_distances_df.values
        if combined_items_distances is None:
            combined_items_distances = np.zeros(numpy_combined_methods_distances.shape)
        combined_items_distances = combined_items_distances + np.power(
            weights[i] * numpy_combined_methods_distances, 2)
    combined_items_distances = pd.DataFrame(np.sqrt(combined_items_distances),
                                            index=combined_methods_distances_df.index,
                                            columns=combined_methods_distances_df.columns)
    combined_items_distances.loc[:, 'composite'] = combined_items_distances.sum(axis=1, skipna=True)
    ranks = []
    for column_name in combined_items_distances.columns:
        combined_items_distances.sort_values(by=column_name, inplace=True)
        column_new_name = column_name.split('_')[0] + '_rank'
        combined_items_distances.loc[:, column_new_name] = \
            combined_items_distances.reset_index().index + 1
        ranks.append(column_new_name)
        combined_items_distances.sort_index(inplace=True)
    if all_ranks:
        display_dataframe = combined_items_distances[ranks]
    else:
        display_dataframe = combined_items_distances[['composite_rank']]
    return display_dataframe


def generate_correlation_scatterplots(api_token, dataframe, folder_name, output_dir='',
                                      api_host=DEFAULT_API_HOST):
    client = GroClient(api_host, api_token)
    logger = client.get_logger()
    folder_path = os.path.join(output_dir, './ranks_csv', folder_name)
    sns.set(style="ticks")
    sns_plot = sns.pairplot(dataframe, diag_kind=None)
    correlation_plot_path = os.path.join(folder_path, 'correlation_plot.png')
    logger.info("Saving scatterplots in {}".format(correlation_plot_path))
    return sns_plot.savefig(correlation_plot_path)


def generate_correlation_matrix(dataframe):
    return dataframe.corr(method='spearman')


def save_to_csv(api_token, dataframe, folder_name, file_name='', output_dir='',
                api_host=DEFAULT_API_HOST):
    """ save the dataframe into csv file called <output_dir>/ranks_csv/ranks.csv """
    client = GroClient(api_host, api_token)
    logger = client.get_logger()
    folder_path = os.path.join(output_dir, './ranks_csv', folder_name)
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    file_path = os.path.join(folder_path, file_name)
    logger.info("\n Saving {} in {} \n".format(file_name, file_path))
    return dataframe.to_csv(file_path)



import os
import unicodecsv
from groclient import GroClient

API_HOST = 'api.gro-intelligence.com'
OUTPUT_FILENAME = 'gro_client_output.csv'
ACCESS_TOKEN = os.environ['GROAPI_TOKEN']


def main():
    client = GroClient(API_HOST, ACCESS_TOKEN)

    selected_entities = {'region_id': 1210,  # Ukraine
                         'item_id': 95,  # Wheat
                         'metric_id': 570001}  # Area Harvested (area)

    writer = unicodecsv.writer(open(OUTPUT_FILENAME, 'wb'))

    # Get what possible series there are for that combination of selections
    for data_series in client.get_data_series(**selected_entities):

        # Add a time range restriction to your data request
        # (Optional - otherwise get all points)
        data_series['start_date'] = '2000-01-01'
        data_series['end_date'] = '2012-12-31'

        for point in client.get_data_points(**data_series):
            writer.writerow([
                point['start_date'],
                point['end_date'],
                point['value'],
                client.lookup_unit_abbreviation(point['unit_id'])
            ])


if __name__ == "__main__":
    main()



"""Search example

Use case: "What data does Gro have available for x?"

This example shows how to query the API for sources matching a search term and
discover what data series are available for a selected source.

Note that the search functions work with any of the following entity types:
 - metrics
 - items
 - regions
 - sources (shown here)

See Also
--------
groclient.GroClient.search()
groclient.GroClient.search_and_lookup()
groclient.GroClient.get_data_series()
https://github.com/gro-intelligence/api-client/wiki/Entities-Definition
https://github.com/gro-intelligence/api-client/wiki/Data-Series-Definition

"""

import os
from groclient import GroClient

API_HOST = 'api.gro-intelligence.com'
ACCESS_TOKEN = os.environ['GROAPI_TOKEN']

def main():
    client = GroClient(API_HOST, ACCESS_TOKEN)

    # ===================
    # | client.search() |
    # ===================
    # Returns a list of ids, ordered by relevance to your given search term
    # Note that you can search across metrics, items, regions, and sources.

    print('client.search()')
    print(client.search('metrics', 'Exports')[0]) # { 'id': 125 }
    print(client.search('items', 'Wheat')[0]) # { 'id': 95 }
    print(client.search('regions', 'India')[0]) # { 'id': 1094 }
    print(client.search('sources', 'USDA NASS')[0]) # { 'id': 29 }

    # ==============================
    # | client.search_and_lookup() |
    # ==============================
    # Helper function to use the client.lookup() function on each search result
    # and see more details about the result.
    # Returns a generator, which yields one search result at a time. Use the
    # next() method to get the first result:
    print('\nclient.search_and_lookup()')
    print(next(client.search_and_lookup('metrics', 'Export Value')))
    # {'id': 10000, 'contains': [10065, 11078], 'name': 'Export Value',
    #  'definition': 'The value of exports, or goods that have been sent to a \
    #  foreign country for sale. Data is mostly reported as free-on-board, \
    #  which includes the cost of delivering the goods to a designated \
    #  delivery vessel; exports of a good may not necessarily equal imports \
    #  for the partner region, since imports and exports are measured \
    #  differently by different governments.'}

    print(next(client.search_and_lookup('items', 'Wheat')))
    # {'id': 95, 'contains': [3595, 5772], 'name': 'Wheat',
    #  'definition': "Cereals within the genus <i>Triticum</i>, which is one \
    #  of the world's most popular and widely cultivated grain crops. Data \
    #  primarily covers common and durum wheat, as well as spelt."}

    print(next(client.search_and_lookup('regions', 'India')))
    # {'id': 1094, 'contains': [11187, 11190, 11174, 11197, 11188, 11200,
    #  11204, 11186, 11180, 11177, 11207, 11201, 11173, 11178, 11195, 11194,
    #  11183, 11199, 11203, 11202, 11193, 11181, 13475, 11196, 11185, 11175,
    #  11198, 11192, 11179, 11191, 11189, 11176, 11182, 11205, 11184, 11206],
    #  'name': 'India', 'level': 3, 'latitude': 22.8838, 'longitude': 79.6201}

    print(next(client.search_and_lookup('sources', 'USDA NASS')))
    # {'id': 29, 'name': 'USDA NASS Animals', 'longName': 'USDA National \
    #  Agricultural Statistics Database', 'metaType': 'data_series',
    #  'sourceLag': {'annual': '4m15d', 'weekly': '4d', 'monthly': '1m10d'},
    #  'historicalStartDate': '1866-12-01T00:00:00.000Z',
    #  'description': 'The National Agricultural Statistics Service is an arm \
    #  of the USDA and one of its primary intelligence- and data-gathering \
    #  units. The database provides updates almost daily on livestock, crops, \
    #  demographics, economics, and environmental indicators. Metrics covered \
    #  include production, yield, area harvested, price, inputs, stocks, etc. \
    #  The granularity is mostly internal US data and goes back as far as \
    #  1850.', 'resolution': 'District', 'regionalCoverage': 'United States',
    #  'language': 'English', 'fileFormat': 'CSV'}

    # ==========================
    # | client.get_data_series |
    # ==========================
    # Once you have identified one or more entities of interest, you can see
    # what data series are available for those entities using the
    # client.get_data_series() function.

    # The normal process of data discovery using the API would be to look up
    # items and/or regions of interest first. i.e. if you know you are
    # interested in United States Corn data. Then you can see what metrics are
    # available for that item and region: production, exports, prices, etc.
    # For example:
    print('\nclient.get_data_series() Part 1: Search by item/region')
    # First look up the item/region of interest as seen in the above examples.
    # We just need the id number, so we will use search(). search_and_lookup()
    # would also work.
    corn = client.search('items', 'corn')[0]
    united_states = client.search('regions', 'united states')[0]
    # Now we can use client.get_data_series() to see what data series exist:
    data_series_list = client.get_data_series(**{
        'item_id': corn['id'],
        'region_id': united_states['id']
    })
    print('There are', len(data_series_list), 'different US Corn data series')
    unique_metrics = set(data_series['metric_name'] for data_series in data_series_list)
    print('Unique metrics:', len(unique_metrics))
    unique_sources = set(data_series['source_name'] for data_series in data_series_list)
    print('Unique sources:', len(unique_sources))

    # If you are interested in a particular source, you can also start there
    # and see what data series exist for it. One frequently asked question is
    # how to see what items/regions Gro publishes yield models for. Here is how
    # one would find out programmatically:
    print('\nclient.get_data_series() Part 2: search by source')
    # Gro publishes its own yield model values under the "Gro Yield Model"
    # source, which is treated as its own source just like any other, and you
    # can find it in the same manner:
    gro_yield_model = client.search('sources', 'Gro Yield Model')[0]
    # Now we can use client.get_data_series() to see what data series exist
    # under that source.
    data_series_list = client.get_data_series(**{
        'source_id': gro_yield_model['id']
    })
    print('There are', len(data_series_list), 'different Gro Yield Model data series')
    # There are thousands of data series in data_series_list since there are
    # many different regions. Let's just check the unique items there are yield
    # models for:
    unique_items = set(data_series['item_name'] for data_series in data_series_list)
    for item in unique_items:
        print(item)
    # Winter wheat
    # Soybeans
    # Hard red winter wheat
    # Corn
    # Wheat

if __name__ == "__main__":
    main()


import argparse
import unicodecsv as csv
from api.client.samples.similar_regions.region_properties import region_properties
from api.client.samples.similar_regions.similar_region import SimilarRegion


def main():
    parser = argparse.ArgumentParser(description="Gro Similar Regions")
    parser.add_argument("--region_id", required=True, type=int, 
            help="region id of the region to which you want to find similar regions")
    parser.add_argument("--csv_output", action='store_true', help="save the output to a csv with filename 'region_id.csv'")
    parser.add_argument("--region_level", default=None, type=int, help="which region level to find similar regions at (3 = country, 4 = province, 5 = district)")
    parser.add_argument("--number_of_regions", default=10, type=int, help="total number of regions to query from the region data set")
    parser.add_argument("--data_dir", default=None, type=str, help="directory to store cached data in. by default in a tmp dir or in a special directory in local dir")
    parser.add_argument("--no_download", action='store_true', help="don't download any new region data; use only cached data")
    args = parser.parse_args()

    if args.csv_output:
        level_suffix = "" if not args.region_level else "_level_" + str(args.region_level)
        f = open(str(args.region_id) + level_suffix + ".csv", 'wb')
        csv_writer = None

    sim = SimilarRegion(region_properties, data_dir=args.data_dir, no_download=args.no_download)
    for output in sim.similar_to(args.region_id, args.number_of_regions, args.region_level):
        if args.csv_output:
            if not csv_writer:
                csv_writer = csv.DictWriter(f, fieldnames= output.keys(),
                                            delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                csv_writer.writeheader()
            csv_writer.writerow(output)
        else:
            print(output)


if __name__ == "__main__":
    main()



from __future__ import print_function
import functools
import itertools
import time
import json

# Python3 support
try:
    # Python3
    from urllib.parse import urlencode
except ImportError:
    # Python2
    from urllib import urlencode

from groclient import cfg, lib
from groclient.constants import DATA_SERIES_UNIQUE_TYPES_ID, ENTITY_KEY_TO_TYPE
from groclient.utils import intersect, zip_selections, dict_unnest
from groclient.lib import APIError

import pandas
from tornado import gen
from tornado.escape import json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from tornado.ioloop import IOLoop
from tornado.queues import Queue


class BatchError(APIError):
    """Replicate the APIError interface given a Tornado HTTPError."""

    def __init__(self, response, retry_count, url, params):
        self.response = response
        self.retry_count = retry_count
        self.url = url
        self.params = params
        self.status_code = (
            self.response.code if hasattr(self.response, "code") else None
        )
        try:
            json_content = json_decode(self.response.body)
            # 'error' should be something like 'Not Found' or 'Bad Request'
            self.message = json_content.get("error", "")
            # Some error responses give additional info.
            # For example, a 400 Bad Request might say "metricId is required"
            if "message" in json_content:
                self.message += ": {}".format(json_content["message"])
        except Exception:
            # If the error message can't be parsed, fall back to a generic "giving up" message.
            self.message = "Giving up on {} after {} {}: {}".format(
                self.url,
                self.retry_count,
                "retry" if self.retry_count == 1 else "retries",
                response,
            )


class GroClient(object):
    """API client with stateful authentication for lib functions and extra convenience methods."""

    def __init__(self, api_host, access_token):
        self.api_host = api_host
        self.access_token = access_token
        self._logger = lib.get_default_logger()
        self._data_series_list = set()  # all that have been added
        self._data_series_queue = []  # added but not loaded in data frame
        self._data_frame = pandas.DataFrame()
        try:
            self._async_http_client = AsyncHTTPClient()
            self._ioloop = IOLoop()
        except Exception as e:
            self._logger.warning(
                "Unable to initialize an event loop. Async methods disabled."
            )
            self._async_http_client = None
            self._ioloop = None

    def __del__(self):
        self._async_http_client.close()
        self._ioloop.stop()
        self._ioloop.close()

    def get_logger(self):
        return self._logger

    @gen.coroutine
    def async_get_data(self, url, headers, params=None):
        base_log_record = dict(route=url, params=params)

        def log_request(start_time, retry_count, msg, status_code):
            elapsed_time = time.time() - start_time
            log_record = dict(base_log_record)
            log_record["elapsed_time_in_ms"] = 1000 * elapsed_time
            log_record["retry_count"] = retry_count
            log_record["status_code"] = status_code
            if status_code == 200:
                self._logger.debug(msg, extra=log_record)
            else:
                self._logger.warning(msg, extra=log_record)

        """General 'make api request' function.

        Assigns headers and builds in retries and logging.
        """
        self._logger.debug(url)

        # append version info
        headers.update(lib.get_version_info())

        # Initialize to -1 so first attempt will be retry 0
        retry_count = -1
        while retry_count <= cfg.MAX_RETRIES:
            retry_count += 1
            start_time = time.time()
            http_request = HTTPRequest(
                "{url}?{params}".format(url=url, params=urlencode(params)),
                method="GET",
                headers=headers,
                request_timeout=cfg.TIMEOUT,
                connect_timeout=cfg.TIMEOUT,
            )
            try:
                try:
                    response = yield self._async_http_client.fetch(http_request)
                    status_code = response.code
                except HTTPError as e:
                    # Catch non-200 codes that aren't errors
                    status_code = e.code if hasattr(e, "code") else None
                    if status_code in [204, 206]:
                        log_msg = {204: "No Content", 206: "Partial Content"}[
                            status_code
                        ]
                        response = e.response
                        log_request(start_time, retry_count, log_msg, status_code)
                        # Do not retry.
                    elif status_code == 301:
                        redirected_ids = json.loads(e.response.body.decode("utf-8"))[
                            "data"
                        ][0]
                        new_params = lib.redirect(params, redirected_ids)
                        log_request(
                            start_time,
                            retry_count,
                            "Redirecting {} to {}".format(params, new_params),
                            status_code,
                        )
                        params = new_params
                        continue  # retry
                    else:  # Otherwise, propagate to error handling
                        raise e
            except Exception as e:
                # HTTPError raised when there's a non-200 status code
                # socket.gaio error raised when there's a connection error
                response = e.response if hasattr(e, "response") else e
                status_code = e.code if hasattr(e, "code") else None
                error_msg = (
                    e.response.error
                    if (hasattr(e, "response") and hasattr(e.response, "error"))
                    else e
                )
                log_request(start_time, retry_count, error_msg, status_code)
                if status_code in [429, 500, 503, 504]:
                    # First retry is immediate.
                    # After that, exponential backoff before retrying.
                    if retry_count > 0:
                        time.sleep(2 ** retry_count)
                    continue
                elif status_code in [400, 401, 402, 404]:
                    break  # Do not retry. Go right to raising an Exception.

            # Request was successful
            log_request(start_time, retry_count, "OK", status_code)
            raise gen.Return(
                json_decode(response.body) if hasattr(response, "body") else None
            )

        # Retries failed. Raise exception
        raise BatchError(response, retry_count, url, params)

    def batch_async_queue(self, func, batched_args, output_list, map_result):
        """Asynchronously call func.

        Parameters
        ----------
        func : function
            The function to be batched. Typically a Client method.
        batched_args : list of dicts
            Inputs
        output_list : any, optional
            A custom accumulator to use in map_result. For example: may pass in a non-empty list
            to append results to it, or may pass in a pandas dataframe, etc. By default, is a list
            of n 0s, where n is the length of batched_args.
        map_result : function, optional
            Function to apply changes to individual requests' responses before returning. Must
            return an accumulator, like a map() function.
            Takes 4 params:
            1. the index in batched_args
            2. the element from batched_args
            3. the result from that input
            4. `output_list`. The accumulator of all results

        """
        assert (
            type(batched_args) is list
        ), "Only argument to a batch async decorated function should be a \
            list of a list of the individual non-keyword arguments being \
            passed to the original function."

        # Wrap output_list in an object so it can be modified within inner functions' scope
        # In Python 3, can accomplish the same thing with `nonlocal` keyword.
        output_data = {}
        if output_list is None:
            output_data["result"] = [0] * len(batched_args)
        else:
            output_data["result"] = output_list

        if not map_result:
            # Default map_result function separates output by index of the query. For example:
            # batched_args: [exports of corn, exports of soybeans]
            # accumulator: [[corn datapoint, corn datapoint],
            #               [soybean data point, soybean data point]]
            def map_result(idx, query, response, accumulator):
                accumulator[idx] = response
                return accumulator

        q = Queue()

        @gen.coroutine
        def consumer():
            """Execute func on all items in queue asynchronously."""
            while q.qsize():
                try:
                    idx, item = q.get().result()
                    self._logger.debug("Doing work on {}".format(idx))
                    if type(item) is dict:
                        # Assume that dict types should be unpacked as kwargs
                        result = yield func(**item)
                    elif type(item) is list:
                        # Assume that list types should be unpacked as positional args
                        result = yield func(*item)
                    else:
                        result = yield func(item)
                    output_data["result"] = map_result(
                        idx, item, result, output_data["result"]
                    )
                    self._logger.debug("Done with {}".format(idx))
                    q.task_done()
                except Exception:
                    # Cease processing
                    # IOLoop raises "Operation timed out after None seconds"
                    self._ioloop.stop()
                    self._ioloop.close()

        def producer():
            """Immediately enqueue the whole batch of requests."""
            lasttime = time.time()
            for idx, item in enumerate(batched_args):
                q.put((idx, item))
            elapsed = time.time() - lasttime
            self._logger.info("Queued {} requests in {}".format(q.qsize(), elapsed))

        @gen.coroutine
        def main():
            # Start consumer without waiting (since it never finishes).
            for i in range(cfg.MAX_QUERIES_PER_SECOND):
                self._ioloop.spawn_callback(consumer)
            producer()  # Wait for producer to put all tasks.
            yield q.join()  # Wait for consumer to finish all tasks.

        self._ioloop.run_sync(main)
        return output_data["result"]

    # TODO: deprecate  the following  two methods, standardize  on one
    # approach with get_data_points and get_df
    @gen.coroutine
    def get_data_points_generator(self, **selection):
        headers = {"authorization": "Bearer " + self.access_token}
        url = "/".join(["https:", "", self.api_host, "v2/data"])
        params = lib.get_data_call_params(**selection)
        try:
            list_of_series_points = yield self.async_get_data(url, headers, params)
            include_historical = selection.get("include_historical", True)
            points = lib.list_of_series_to_single_series(
                list_of_series_points, False, include_historical
            )
            raise gen.Return(points)
        except BatchError as b:
            raise gen.Return(b)

    def batch_async_get_data_points(
        self, batched_args, output_list=None, map_result=None
    ):
        """Make many :meth:`~get_data_points` requests asynchronously.

        Parameters
        ----------
        batched_args : list of dicts
            Each dict should be a `selections` object like would be passed to
            :meth:`~get_data_points`.

            Example::

                input_list = [
                    {'metric_id': 860032, 'item_id': 274, 'region_id': 1215, 'frequency_id': 9, 'source_id': 2},
                    {'metric_id': 860032, 'item_id': 270, 'region_id': 1215, 'frequency_id': 9, 'source_id': 2}
                ]

        output_list : any, optional
            A custom accumulator to use in map_result. For example: may pass in a non-empty list
            to append results to it, or may pass in a pandas dataframe, etc. By default, is a list
            of n 0s, where n is the length of batched_args.
        map_result : function, optional
            Function to apply changes to individual requests' responses before returning.
            Takes 4 params:
            1. the index in batched_args
            2. the element from batched_args
            3. the result from that input
            4. `output_list`. The accumulator of all results

            Example::

                output_list = []

                # Merge all responses into a single list
                def map_response(inputIndex, inputObject, response, output_list):
                    output_list += response
                    return output_list

                batch_output = client.batch_async_get_data_points(input_list,
                                                                  output_list=output_list,
                                                                  map_result=map_response)

        Returns
        -------
        any
            By default, returns a list of lists of data points. Likely either objects or lists of
            dictionaries. If using a custom map_result function, can return any type.

            Example of the default output format::

                [
                    [
                        {'metric_id': 1, 'item_id': 2, 'start_date': 2000-01-01, 'value': 41, ...},
                        {'metric_id': 1, 'item_id': 2, 'start_date': 2001-01-01, 'value': 39, ...},
                        {'metric_id': 1, 'item_id': 2, 'start_date': 2002-01-01, 'value': 50, ...},
                    ],
                    [
                        {'metric_id': 1, 'item_id': 6, 'start_date': 2000-01-01, 'value': 12, ...},
                        {'metric_id': 1, 'item_id': 6, 'start_date': 2001-01-01, 'value': 13, ...},
                        {'metric_id': 1, 'item_id': 6, 'start_date': 2002-01-01, 'value': 4, ...},
                    ],
                ]

        """
        return self.batch_async_queue(
            self.get_data_points_generator, batched_args, output_list, map_result
        )

    @gen.coroutine
    def async_rank_series_by_source(self, *selections_list):
        """Get all sources, in ranked order, for a given selection."""
        response = self.rank_series_by_source(selections_list)
        raise gen.Return(list(response))

    def batch_async_rank_series_by_source(
        self, batched_args, output_list=None, map_result=None
    ):
        """Perform multiple rank_series_by_source requests asynchronously.

        Parameters
        ----------
        batched_args : list of lists of dicts
            See :meth:`~.rank_series_by_source` `selections_list`. A list of those lists.

        """
        return self.batch_async_queue(
            self.async_rank_series_by_source, batched_args, output_list, map_result
        )

    def get_available(self, entity_type):
        """List the first 5000 available entities of the given type.

        Parameters
        ----------
        entity_type : {'metrics', 'items', 'regions'}

        Returns
        -------
        data : list of dicts

            Example::

                [ { 'id': 0, 'contains': [1, 2, 3], 'name': 'World', 'level': 1},
                  { 'id': 1, 'contains': [4, 5, 6], 'name': 'Asia', 'level': 2},
                ... ]

        """
        return lib.get_available(self.access_token, self.api_host, entity_type)

    def list_available(self, selected_entities):
        """List available entities given some selected entities.

        Given one or more selections, return entities combinations that have
        data for the given selections.

        Parameters
        ----------
        selected_entities : dict

            Example::

                { 'metric_id': 123, 'item_id': 456, 'source_id': 7 }

            Keys may include: metric_id, item_id, region_id, partner_region_id,
            source_id, frequency_id

        Returns
        -------
        list of dicts

            Example::

                [ { 'metric_id': 11078, 'metric_name': 'Export Value (currency)',
                    'item_id': 274, 'item_name': 'Corn',
                    'region_id': 1215, 'region_name': 'United States',
                    'source_id': 15, 'source_name': 'USDA GATS' },
                  { ... },
                ... ]

        """
        return lib.list_available(self.access_token, self.api_host, selected_entities)

    def lookup(self, entity_type, entity_ids):
        """Retrieve details about a given id or list of ids of type entity_type.

        https://developers.gro-intelligence.com/gro-ontology.html

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions', 'frequencies', 'sources', 'units' }
        entity_ids : int or list of ints

        Returns
        -------
        dict or dict of dicts
            A dict with entity details is returned if an integer is given for entity_ids.
            A dict of dicts with entity details, keyed by id, is returned if a list of integers is
            given for entity_ids.

            Example::

                { 'id': 274,
                  'contains': [779, 780, ...]
                  'name': 'Corn',
                  'definition': 'The seeds of the widely cultivated corn plant <i>Zea mays</i>,'
                                ' which is one of the world\'s most popular grains.' }

            Example::

                {   '274': {
                        'id': 274,
                        'contains': [779, 780, ...],
                        'belongsTo': [4138, 8830, ...],
                        'name': 'Corn',
                        'definition': 'The seeds of the widely cultivated corn plant'
                                      ' <i>Zea mays</i>, which is one of the world\'s most popular'
                                      ' grains.'
                    },
                    '270': {
                        'id': 270,
                        'contains': [1737, 7401, ...],
                        'belongsTo': [8830, 9053, ...],
                        'name': 'Soybeans',
                        'definition': 'The seeds and harvested crops of plants belonging to the'
                                      ' species <i>Glycine max</i> that are used in the production'
                                      ' of oil and both human and livestock consumption.'
                    }
                }

        """
        return lib.lookup(self.access_token, self.api_host, entity_type, entity_ids)

    def lookup_unit_abbreviation(self, unit_id):
        return self.lookup("units", unit_id)["abbreviation"]

    def get_allowed_units(self, metric_id, item_id=None):
        """Get a list of unit that can be used with the given metric (and
        optionally, item).

        Parameters
        ----------
        metric_id: int
        item_id: int, optional.


        Returns
        -------
        list of unit ids

        """
        return lib.get_allowed_units(
            self.access_token, self.api_host, metric_id, item_id
        )

    def get_data_series(self, **selection):
        """Get available data series for the given selections.

        https://developers.gro-intelligence.com/data-series-definition.html

        Parameters
        ----------
        metric_id : integer, optional
        item_id : integer, optional
        region_id : integer, optional
        partner_region_id : integer, optional
        source_id : integer, optional
        frequency_id : integer, optional

        Returns
        -------
        list of dicts

            Example::

                [{ 'metric_id': 2020032, 'metric_name': 'Seed Use',
                   'item_id': 274, 'item_name': 'Corn',
                   'region_id': 1215, 'region_name': 'United States',
                   'source_id': 24, 'source_name': 'USDA FEEDGRAINS',
                   'frequency_id': 7,
                   'start_date': '1975-03-01T00:00:00.000Z',
                   'end_date': '2018-05-31T00:00:00.000Z'
                 }, { ... }, ... ]

        """
        return lib.get_data_series(self.access_token, self.api_host, **selection)

    def search(self, entity_type, search_terms):
        """Search for the given search term. Better matches appear first.

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions', 'sources' }
        search_terms : string

        Returns
        -------
        list of dicts

            Example::

                [{'id': 5604}, {'id': 10204}, {'id': 10210}, ....]

        """
        return lib.search(self.access_token, self.api_host, entity_type, search_terms)

    def search_and_lookup(self, entity_type, search_terms, num_results=10):
        """Search for the given search terms and look up their details.

        For each result, yield a dict of the entity and it's properties.

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions', 'sources' }
        search_terms : string
        num_results: int
            Maximum number of results to return. Defaults to 10.

        Yields
        ------
        dict
            Result from :meth:`~.search` passed to :meth:`~.lookup` to get additional details.

            Example::

                { 'id': 274,
                  'contains': [779, 780, ...],
                  'name': 'Corn',
                  'definition': 'The seeds of the widely cultivated...' }

            See output of :meth:`~.lookup`. Note that as with :meth:`~.search`, the first result is
            the best match for the given search term(s).

        """
        return lib.search_and_lookup(
            self.access_token, self.api_host, entity_type, search_terms, num_results
        )

    def lookup_belongs(self, entity_type, entity_id):
        """Look up details of entities containing the given entity.

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions' }
        entity_id : int

        Yields
        ------
        dict
            Result of :meth:`~.lookup` on each entity the given entity belongs to.

            For example: For the region 'United States', one yielded result will be for
            'North America.' The format of which matches the output of :meth:`~.lookup`::

                { 'id': 15,
                  'contains': [ 1008, 1009, 1012, 1215, ... ],
                  'name': 'North America',
                  'level': 2 }

        """
        return lib.lookup_belongs(
            self.access_token, self.api_host, entity_type, entity_id
        )

    def rank_series_by_source(self, selections_list):
        """Given a list of series selections, for each unique combination excluding source, expand
        to all available sources and return them in ranked order. The order corresponds to how well
        that source covers the selection (metrics, items, regions, and time range and frequency).

        Parameters
        ----------
        selections_list : list of dicts
            See the output of :meth:`~.get_data_series`.

        Yields
        ------
        dict
            The input selections_list, expanded out to each possible source, ordered by coverage.

        """
        return lib.rank_series_by_source(
            self.access_token, self.api_host, selections_list
        )

    def get_geo_centre(self, region_id):
        """Given a region ID, return the geographic centre in degrees lat/lon.

        Parameters
        ----------
        region_id : integer

        Returns
        -------
        list of dicts

            Example::

                [{'centre': [ 39.8333, -98.5855 ], 'regionId': 1215, 'regionName': 'United States'}]
        """
        return lib.get_geo_centre(self.access_token, self.api_host, region_id)

    def get_geojsons(self, region_id, descendant_level=None, zoom_level=7):
        """Given a region ID, return shape information in geojson, for the
        region and all its descendants at the given level (if specified).

        Parameters
        ----------
        region_id : integer
        descendant_level : integer, admin region level (2, 3, 4 or 5)
        zoom_level : integer, optional(allow 1-8)
            Valid if include_geojson equals True. If zoom level is specified and it is less than 6,
            simplified shapefile will be returned. Otherwise, detailed shapefile will be used by default.

        Returns
        -------
        list of dicts

            Example::

                [{  'centre': [ 39.8333, -98.5855 ],
                    'regionId': 1215,
                    'regionName': 'United States',
                    u'geojson': u'{"type":"GeometryCollection","geometries":[{"type":"MultiPolygon","coordinates":[[[[-155.651382446,20.1647224430001], ...]]]}]}'
                }]

        """
        return lib.get_geojsons(
            self.access_token, self.api_host, region_id, descendant_level, zoom_level)

    def get_geojson(self, region_id, zoom_level=7):
        """Given a region ID, return shape information in geojson.

        Parameters
        ----------
        region_id : integer
        zoom_level : integer, optional(allow 1-8)
            Valid if include_geojson equals True. If zoom level is specified and it is less than 6,
            simplified shapefile will be returned. Otherwise, detailed shapefile will be used by default.

        Returns
        -------
        a geojson object or None

            Example::

                { 'type': 'GeometryCollection',
                'geometries': [{'type': 'MultiPolygon',
                                'coordinates': [[[[-38.394, -4.225], ...]]]}, ...]}

        """
        return lib.get_geojson(self.access_token, self.api_host, region_id, zoom_level)

    def get_descendant(
        self,
        entity_type,
        entity_id,
        distance=None,
        include_details=True,
    ):
        """Given an item, metric or region, returns all its descendants i.e. 
        entities that are "contained" in the given entity

        Similar to :meth:~.get_descendant_regions, but also works on items and metrics. This method has 
        a distance parameter (which returns all nested child entities) instead of a descendant_level 
        parameter (which only returns child entities at a given depth/level).

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions' }
        entity_id : integer
        distance: integer, optional
            Return all entity contained to entity_id at maximum distance.
            If not provided, get all descendants.
        include_details : boolean, optional
            True by default. Will perform a lookup() on each descendant  to find name,
            definition, etc. If this option is set to False, only ids of descendant
            entities will be returned, which makes execution significantly faster.

        Returns
        -------
        list of dicts

            Example::

                [{
                    'id': 134,
                    'name': 'Cattle hides, wet-salted',
                    'definition': 'Hides and skins of domesticated cattle-animals ...',
                } , {
                    'id': 382,
                    'name': 'Calf skins, wet-salted',
                    'definition': 'Wet-salted hides and skins of calves-animals of ...'
                }, ...]

            See output of :meth:`~.lookup`

        """
        return lib.get_descendant(
            self.access_token,
            self.api_host,
            entity_type,
            entity_id,
            distance,
            include_details,
        )


    def get_descendant_regions(
        self,
        region_id,
        descendant_level=None,
        include_historical=True,
        include_details=True,
    ):
        """Look up details of all regions of the given level contained by a region.

        Given any region by id, get all the descendant regions that are of the specified level.

        Parameters
        ----------
        region_id : integer
        descendant_level : integer, optional
            The region level of interest. See REGION_LEVELS constant. If not provided, get all
            descendants.
        include_historical : boolean, optional
            True by default. If False is specified, regions that only exist in historical data
            (e.g. the Soviet Union) will be excluded.
        include_details : boolean, optional
            True by default. Will perform a lookup() on each descendant region to find name,
            latitude, longitude, etc. If this option is set to False, only ids of descendant
            regions will be returned, which makes execution significantly faster.

        Returns
        -------
        list of dicts

            Example::

                [{
                    'id': 13100,
                    'contains': [139839, 139857, ...],
                    'name': 'Wisconsin',
                    'level': 4
                } , {
                    'id': 13101,
                    'contains': [139891, 139890, ...],
                    'name': 'Wyoming',
                    'level': 4
                }, ...]

            See output of :meth:`~.lookup`

        """
        return lib.get_descendant_regions(
            self.access_token,
            self.api_host,
            region_id,
            descendant_level,
            include_historical,
            include_details,
        )

    def get_available_timefrequency(self, **selection):
        """Given a selection, return a list of frequencies and time ranges.
        The results are ordered by coverage-optimized ranking.

        Parameters
        ----------
        metric_id : integer, optional
        item_id : integer, optional
        region_id : integer, optional
        partner_region_id : integer, optional

        Returns
        -------
        list of dicts

            Example::

                 [{
                    'start_date': '2000-02-18T00:00:00.000Z',
                    'frequency_id': 3,
                    'end_date': '2020-03-12T00:00:00.000Z',
                    'name': '8-day'
                  }, {
                    'start_date': '2019-09-02T00:00:00.000Z',
                    'frequency_id': 1,
                    'end_date': '2020-03-09T00:00:00.000Z',
                    'name': u'daily'}, ... ]
        """
        return lib.get_available_timefrequency(
            self.access_token, self.api_host, **selection
        )

    def get_top(self, entity_type, num_results=5, **selection):
        """Find the data series with the highest cumulative value for the given time range.

        Examples::

            # To get FAO's top 5 corn-producing countries of all time:
            client.get_top('regions', metric_id=860032, item_id=274, frequency_id=9, source_id=2)

            # To get FAO's top 5 corn-producing countries of 2014:
            client.get_top('regions', metric_id=860032, item_id=274, frequency_id=9, source_id=2,
                           start_date='2014-01-01', end_date='2014-12-31')

            # To get the United States' top 15 exports in the decade of 2010-2019:
            client.get_top('items', num_results=15, metric_id=20032, region_id=1215, frequency_id=9,
                           source_id=2, start_date='2010-01-01', end_date='2019-12-31')

        Parameters
        ----------
        entity_type : { 'items', 'regions' }
            The entity type to rank, all other selections being the same. Only items and regions
            are rankable at this time.
        num_results : integer, optional
            How many data series to rank. Top 5 by default.
        metric_id : integer
        item_id : integer
            Required if requesting top regions. Disallowed if requesting top items.
        region_id : integer
            Required if requesting top items. Disallowed if requesting top regions.
        partner_region_id : integer, optional
        frequency_id : integer
        source_id : integer
        start_date : string, optional
            If not provided, the cumulative value used for ranking will include data points as far
            back as the source provides.
        end_date : string, optional

        Returns
        -------
        list of dicts

            Example::

                [
                    {'metricId': 860032, 'itemId': 274, 'regionId': 1215, 'frequencyId': 9,
                     'sourceId': 2, 'value': 400, 'unitId': 14},
                    {'metricId': 860032, 'itemId': 274, 'regionId': 1215, 'frequencyId': 9,
                     'sourceId': 2, 'value': 395, 'unitId': 14},
                    {'metricId': 860032, 'itemId': 274, 'regionId': 1215, 'frequencyId': 9,
                     'sourceId': 2, 'value': 12, 'unitId': 14},
                ]

            Along with the series attributes, value and unit are also given for the total cumulative
            value the series are ranked by. You may then use the results to call
            :meth:`~.get_data_points` to get the individual time series points.
        """
        return lib.get_top(
            self.access_token, self.api_host, entity_type, num_results, **selection
        )

    def get_df(self, show_revisions=False, index_by_series=False):
        """Call :meth:`~.get_data_points` for each saved data series and return as a combined
        dataframe.

        Note you must have first called either :meth:`~.add_data_series` or
        :meth:`~.add_single_data_series` to save data series into the GroClient's data_series_list.
        You can inspect the client's saved list using :meth:`~.get_data_series_list`.

        Returns
        -------
        pandas.DataFrame
            The results to :meth:`~.get_data_points` for all the saved series, appended together
            into a single dataframe.
            See https://developers.gro-intelligence.com/data-point-definition.html
            If index_by_series is set, the dataframe is indexed by series.
            See https://developers.gro-intelligence.com/data-series-definition.html
        """
        while self._data_series_queue:
            data_series = self._data_series_queue.pop()
            if show_revisions:
                data_series["show_revisions"] = True
            self.add_points_to_df(
                None, data_series, self.get_data_points(**data_series)
            )
        if index_by_series and not self._data_frame.empty:
            columns = intersect(DATA_SERIES_UNIQUE_TYPES_ID, self._data_frame.columns)
            if len(columns) > 0:
                indexed_df = self._data_frame.set_index(columns)
                indexed_df.index.set_names(DATA_SERIES_UNIQUE_TYPES_ID, inplace=True)
                return indexed_df.sort_index()
        return self._data_frame

    def async_get_df(self):
        self.batch_async_get_data_points(
            self._data_series_queue,
            output_list=self._data_frame,
            map_result=self.add_points_to_df,
        )
        return self._data_frame

    def add_points_to_df(self, index, data_series, data_points, *args):
        """Add the given datapoints to a pandas dataframe.

        Parameters:
        -----------
        index : unused
        data_series : dict
        data_points : list of dicts

        """

        tmp = pandas.DataFrame(data=[dict_unnest(point) for point in data_points])
        if tmp.empty:
            return
        # get_data_points response doesn't include the
        # source_id. We add it as a column, in case we have
        # several selections series which differ only by source id.
        tmp["source_id"] = data_series["source_id"]
        # tmp should always have end_date/start_date/reporting_date as columns if not empty
        tmp.end_date = pandas.to_datetime(tmp.end_date)
        tmp.start_date = pandas.to_datetime(tmp.start_date)
        tmp.reporting_date = pandas.to_datetime(tmp.reporting_date)

        if self._data_frame.empty:
            self._data_frame = tmp
        else:
            self._data_frame = pandas.concat([self._data_frame, tmp])

    def get_data_points(self, **selections):
        """Get all the data points for a given selection.

        https://developers.gro-intelligence.com/data-point-definition.html

        Example::

            client.get_data_points(**{'metric_id': 860032,
                                      'item_id': 274,
                                      'region_id': 1215,
                                      'frequency_id': 9,
                                      'source_id': 2,
                                      'start_date': '2017-01-01',
                                      'end_date': '2017-12-31',
                                      'unit_id': 15})

        Returns::

            [{  'start_date': '2017-01-01T00:00:00.000Z',
                'end_date': '2017-12-31T00:00:00.000Z',
                'value': 408913833.8019222, 'unit_id': 15,
                'reporting_date': None,
                'metric_id': 860032, 'item_id': 274, 'region_id': 1215,
                'partner_region_id': 0, 'frequency_id': 9, 'source_id': 2,
                'belongs_to': {
                    'metric_id': 860032,
                    'item_id': 274,
                    'region_id': 1215,
                    'frequency_id': 9,
                    'source_id': 2
                }
            }]

        Note: you can pass the output of :meth:`~.get_data_series` into :meth:`~.get_data_points`
        to check what series exist for some selections and then retrieve the data points for those
        series. See :sample:`quick_start.py` for an example of this.

        :meth:`~.get_data_points` also allows passing a list of ids for metric_id, item_id, and/or
        region_id to get multiple series in a single request. This can be faster if requesting many
        series.

        For example::

            client.get_data_points(**{'metric_id': 860032,
                                      'item_id': 274,
                                      'region_id': [1215,1216],
                                      'frequency_id': 9,
                                      'source_id': 2,
                                      'start_date': '2017-01-01',
                                      'end_date': '2017-12-31',
                                      'unit_id': 15})
        Returns::

            [{  'start_date': '2017-01-01T00:00:00.000Z',
                'end_date': '2017-12-31T00:00:00.000Z',
                'value': 408913833.8019222, 'unit_id': 15,
                'reporting_date': None,
                'metric_id': 860032, 'item_id': 274, 'region_id': 1215,
                'partner_region_id': 0, 'frequency_id': 9, 'source_id': 2,
                'belongs_to': {
                    'metric_id': 860032,
                    'item_id': 274,
                    'region_id': 1215,
                    'frequency_id': 9,
                    'source_id': 2
                }
            }, { 'start_date': '2017-01-01T00:00:00.000Z',
                 'end_date': '2017-12-31T00:00:00.000Z',
                 'value': 340614.19507563586, 'unit_id': 15,
                 'reporting_date': None,
                 'metric_id': 860032, 'item_id': 274, 'region_id': 1216,
                 'partner_region_id': 0, 'frequency_id': 9, 'source_id': 2,
                 'belongs_to': {
                    'metric_id': 860032,
                    'item_id': 274,
                    'region_id': 1216,
                    'frequency_id': 9,
                    'source_id': 2
                 }
            }]

        Parameters
        ----------
        metric_id : integer or list of integers
            How something is measured. e.g. "Export Value" or "Area Harvested"
        item_id : integer or list of integers
            What is being measured. e.g. "Corn" or "Rainfall"
        region_id : integer or list of integers
            Where something is being measured e.g. "United States Corn Belt" or "China"
        partner_region_id : integer or list of integers, optional
            partner_region refers to an interaction between two regions, like trade or
            transportation. For example, for an Export metric, the "region" would be the exporter
            and the "partner_region" would be the importer. For most series, this can be excluded
            or set to 0 ("World") by default.
        source_id : integer
        frequency_id : integer
        unit_id : integer, optional
        start_date : string, optional
            All points with end dates equal to or after this date
        end_date : string, optional
            All points with start dates equal to or before this date
        show_revisions : boolean, optional
            False by default, meaning only the latest value for each period. If true, will return
            all values for a given period, differentiated by the `reporting_date` field.
        insert_null : boolean, optional
            False by default. If True, will include a data point with a None value for each period
            that does not have data.
        at_time : string, optional
            Estimate what data would have been available via Gro at a given time in the past. See
            :sample:`at-time-query-examples.ipynb` for more details.
        include_historical : boolean, optional
            True by default, will include historical regions that are part of your selections

        Returns
        -------
        list of dicts

        """
        data_points = lib.get_data_points(
            self.access_token, self.api_host, **selections
        )
        # Apply unit conversion if a unit is specified
        if "unit_id" in selections:
            return list(
                map(
                    functools.partial(
                        self.convert_unit, target_unit_id=selections["unit_id"]
                    ),
                    data_points,
                )
            )
        # Return data points in input units if not unit is specified
        return data_points

    def GDH(self, gdh_selection, **optional_selections):
        """Wrapper for :meth:`~.get_data_points`. with alternative input and output style.

        The data series selection to retrieve is encoded in a
        'gdh_selection' string of the form
        <metric_id>-<item_id>-<region_id>-<partner_region_id>-<frequency_id>-<source_id>

        For example, client.GDH("860032-274-1231-0-9-14") will get the
        data points for Production of Corn in China from PS&D at an
        annual frequency, e.g.
        for csv_row in client.GDH("860032-274-1231-0-9-14"):
            print csv_row

        Parameters:
        ----------
        gdh_selection: string
        optional_selections: dict, optional
            accepts optional params from :meth:`~.get_data_points`.

        Returns:
        ------
        pandas.DataFrame

            the subset of the main DataFrame :meth:`~.get_df`. with the requested series.

        """

        entity_ids = [int(x) for x in gdh_selection.split("-")]
        selection = zip_selections(entity_ids, **optional_selections)

        self.add_single_data_series(selection)
        try:
            return self.get_df(index_by_series=True).loc[[tuple(entity_ids)], :]
        except KeyError:
            self._logger.warning("GDH returned no data")
            return pandas.DataFrame()

    def get_data_series_list(self):
        """Inspect the current list of saved data series contained in the GroClient.

        For use with :meth:`~.get_df`. Add new data series to the list using
        :meth:`~.add_data_series` and :meth:`~.add_single_data_series`.

        Returns
        -------
        list of dicts
            A list of data_series objects, as returned by :meth:`~.get_data_series`.

        """
        return [dict(data_series_hash) for data_series_hash in self._data_series_list]

    def add_single_data_series(self, data_series):
        """Save a data series object to the GroClient's data_series_list.

        For use with :meth:`~.get_df`.

        Parameters
        ----------
        data_series : dict
            A single data_series object, as returned by :meth:`~.get_data_series` or
            :meth:`~.find_data_series`.
            See https://developers.gro-intelligence.com/data-series-definition.html

        Returns
        -------
        None

        """
        series_hash = frozenset(dict_unnest(data_series).items())
        if series_hash not in self._data_series_list:
            self._data_series_list.add(series_hash)
            # Add a copy of the data series, in case the original is modified
            self._data_series_queue.append(dict(data_series))
            self._logger.info("Added {}".format(data_series))
        else:
            self._logger.debug("Already added: {}".format(data_series))
        return

    def find_data_series(self, result_filter=None, **kwargs):
        """Find data series matching a combination of entities specified by
        name and yield them ranked by coverage.

        Example::

            client.find_data_series(item="Corn",
                                    metric="Futures Open Interest",
                                    region="United States of America")

        will yield a sequence of dictionaries of the form::

            { 'metric_id': 15610005, 'metric_name': 'Futures Open Interest',
              'item_id': 274, 'item_name': 'Corn',
              'region_id': 1215, 'region_name': 'United States',
              'frequency_id': 15, 'source_id': 81,
              'start_date': '1972-03-01T00:00:00.000Z', ...},
            { ... },  ...


        See https://developers.gro-intelligence.com/data-series-definition.html

        :code:`result_filter` can be used to filter entity searches. For example::

            client.find_data_series(item="vegetation",
                                    metric="vegetation indices",
                                    region="Central",
                                    result_filter=lambda r: ('region_id' not in r or
                                                             r['region_id'] == 10393))

        will only consider that particular region, and not the many other regions
        with the same name.

        This method uses :meth:`~.search`, :meth:`~.get_data_series`,
        :meth:`~.get_available_timefrequency` and  :meth:`~.rank_series_by_source`.


        Parameters
        ----------
        metric : string, optional
        item : string, optional
        region : string, optional
        partner_region : string, optional
        start_date : string, optional
            YYYY-MM-DD
        end_date : string, optional
            YYYY-MM-DD
        result_filter: function, optional
            function taking data series selection dict returning boolean

        Yields
        ------
        dict
           A sequence of data series matching the input selections

        See also
        --------
        :meth:`~.get_data_series`

        """
        results = []  # [[('item_id',1),('item_id',2),...],[('metric_id" 1),...],...]
        for kw in kwargs:
            if kwargs.get(kw) is None:
                continue
            id_key = "{}_id".format(kw)
            if id_key in ENTITY_KEY_TO_TYPE:
                type_results = []  # [('item_id',1),('item_id',2),...]
                for search_result in self.search(
                    ENTITY_KEY_TO_TYPE[id_key], kwargs[kw]
                )[: cfg.MAX_RESULT_COMBINATION_DEPTH]:
                    if result_filter is None or result_filter(
                        {id_key: search_result["id"]}
                    ):
                        type_results.append((id_key, search_result["id"]))
                results.append(type_results)
        # Rank by frequency and source, while preserving search ranking in
        # permutations of search results.
        ranking_groups = set()
        for comb in itertools.product(*results):
            for data_series in self.get_data_series(**dict(comb))[
                : cfg.MAX_SERIES_PER_COMB
            ]:
                self._logger.debug("Data series: {}".format(data_series))
                # remove time and frequency to rank them
                data_series.pop("start_date", None)
                data_series.pop("end_date", None)
                data_series.pop("frequency_id", None)
                data_series.pop("frequency_name", None)
                # remove source to rank them
                data_series.pop("source_id", None)
                data_series.pop("source_name", None)
                # metadata is not hashable
                data_series.pop("metadata", None)
                series_hash = frozenset(data_series.items())
                if series_hash not in ranking_groups:
                    ranking_groups.add(series_hash)
                    if kwargs.get("start_date"):
                        data_series["start_date"] = kwargs["start_date"]
                    if kwargs.get("end_date"):
                        data_series["end_date"] = kwargs["end_date"]
                    for tf in self.get_available_timefrequency(**data_series):
                        ds = dict(data_series)
                        ds["frequency_id"] = tf["frequency_id"]
                        for data_series in self.rank_series_by_source([ds]):
                            yield self.get_data_series(**data_series)[0]

    def add_data_series(self, **kwargs):
        """Adds the top result of :meth:`~.find_data_series` to the saved data series list.

        For use with :meth:`~.get_df`.

        Parameters
        ----------
        metric : string, optional
        item : string, optional
        region : string, optional
        partner_region : string, optional
        start_date : string, optional
            YYYY-MM-DD
        end_date : string, optional
            YYYY-MM-DD
        result_filter: function, optional
            function taking data series selection dict returning boolean

        Returns
        -------
        data_series object, as returned by :meth:`~.get_data_series`.
            The data_series that was added or None if none were found.

        See also
        --------
        :meth:`~.get_df`
        :meth:`~.add_single_data_series`
        :meth:`~.find_data_series`

        """
        for the_data_series in self.find_data_series(**kwargs):
            self.add_single_data_series(the_data_series)
            return the_data_series
        return

    ###
    # Discovery shortcuts
    ###
    def search_for_entity(self, entity_type, keywords):
        """Returns the first result of entity_type that matches the given keywords.

        Parameters
        ----------
        entity_type : { 'metrics', 'items', 'regions', 'sources' }
        keywords : string

        Returns
        ----------
        integer
            The id of the first search result

        """
        results = self.search(entity_type, keywords)
        for result in results:
            self._logger.debug(
                "First result, out of {} {}: {}".format(
                    len(results), entity_type, result["id"]
                )
            )
            return result["id"]

    def get_provinces(self, country_name):
        """Given the name of a country, find its provinces.

        Parameters
        ----------
        country_name : string

        Returns
        ----------
        list of dicts

            Example::

                [{
                    'id': 13100,
                    'contains': [139839, 139857, ...],
                    'name': 'Wisconsin',
                    'level': 4
                } , {
                    'id': 13101,
                    'contains': [139891, 139890, ...],
                    'name': 'Wyoming',
                    'level': 4
                }, ...]

            See output of :meth:`~.lookup`

        See Also
        --------
        :meth:`~.get_descendant_regions`

        """
        for region in self.search_and_lookup("regions", country_name):
            if region["level"] == lib.REGION_LEVELS["country"]:
                provinces = self.get_descendant_regions(
                    region["id"], lib.REGION_LEVELS["province"]
                )
                self._logger.debug(
                    "Provinces of {}: {}".format(country_name, provinces)
                )
                return provinces
        return None

    def get_names_for_selection(self, selection):
        """Convert a selection into entity names.

        Parameters:
        -----------
        data_series : dict
            A single data_series object, as returned by get_data_series() or find_data_series().
            See https://github.com/gro-intelligence/api-client/wiki/Data-Series-Definition

        Returns:
        --------
        list of pairs of strings
            [('item', 'Corn'), ('region', 'China') ...]

        """
        return [
            (
                entity_key.split("_")[0],
                self.lookup(ENTITY_KEY_TO_TYPE[entity_key], entity_id)["name"],
            )
            for entity_key, entity_id in selection.items()
        ]

    def convert_unit(self, point, target_unit_id):
        """Convert the data point from one unit to another unit.

        If original or target unit is non-convertible, throw an error.

        Parameters
        ----------
        point : dict
            { value: float, unit_id: integer, ... }
        target_unit_id : integer

        Returns
        -------
        dict

            Example ::

                { value: 14.2, unit_id: 4 }

            unit_id is changed to the target, and value is converted to use the
            new unit_id. Other properties are unchanged.

        """
        if point.get("unit_id") is None or point.get("unit_id") == target_unit_id:
            return point
        from_convert_factor = self.lookup("units", point["unit_id"]).get(
            "baseConvFactor"
        )
        if not from_convert_factor.get("factor"):
            raise Exception("unit_id {} is not convertible".format(point["unit_id"]))
        to_convert_factor = self.lookup("units", target_unit_id).get("baseConvFactor")
        if not to_convert_factor.get("factor"):
            raise Exception("unit_id {} is not convertible".format(target_unit_id))

        if point.get("value") is not None:
            point["value"] = lib.convert_value(point["value"], from_convert_factor, to_convert_factor)
        if point.get("metadata") is not None and point["metadata"].get("conf_interval") is not None:
            point["metadata"]["conf_interval"] = lib.convert_value(point["metadata"]["conf_interval"], from_convert_factor, to_convert_factor)
        point["unit_id"] = target_unit_id
        return point



import argparse
from dateutil.parser import parse
from functools import partial
import os

from groclient import GroClient

from api.client.samples.analogous_years.lib import final_ranks_computation

API_HOST = "api.gro-intelligence.com"


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def valid_date(s):
    """Checks if the provided dates are valid dates in YYYY-MM-DD format"""
    try:
        parse(s)
        return s
    except ValueError:
        msg = "Not a valid date in YYYY-MM-DD format: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


def list_length_validator(list1, list2):
    """Checks if the 2nd list is of the same length as the first and returns the second list
    on success"""
    if len(list1) != len(list2):
        msg = "Mismatch between the number of entries in {} and {}".format(list1, list2)
        raise argparse.ArgumentTypeError(msg)
    else:
        return list2


def check_if_exists(entity_type, entity_value, api_token, api_host=API_HOST):
    """Checks if the Gro-entity_id exists"""
    client = GroClient(api_host, api_token)
    logger = client.get_logger()
    try:
        client.lookup(entity_type, entity_value)
        return entity_value
    except Exception as e:
        message = "Gro-{}_id invalid: '{}'.".format(entity_type, entity_value)
        logger.warning(message)
        raise e


def get_data_series_list(region_id_list, item_id_list, metric_id_list, source_id_list,
                         frequency_id_list, api_token, api_host=API_HOST):
    # checking if the length of the list for metric_id, item_id, source_id and
    # frequency_id match
    item_id_list = list_length_validator(metric_id_list, item_id_list)
    source_id_list = list_length_validator(metric_id_list, source_id_list)
    frequency_id_list = list_length_validator(metric_id_list, frequency_id_list)
    data_series_list = []
    checking = partial(check_if_exists, api_token=api_token)
    for i in range(len(metric_id_list)):
        data_series = {'metric_id': checking('metrics', metric_id_list[i]),
                       'item_id': checking('items', item_id_list[i]),
                       'region_id': checking('regions', region_id_list[0]),
                       'source_id': checking('sources', source_id_list[i]),
                       'frequency_id': checking('frequencies', frequency_id_list[i])}
        data_series_list.append(data_series)
    return data_series_list


def main():
    parser = argparse.ArgumentParser(description='ConGrouent Years')
    parser.add_argument('-m', '--metric_ids', nargs='+', type=int, default=[2100031, 2540047],
                        help='metric_ids separated by spaces')
    parser.add_argument('-i', '--item_ids', nargs='+', type=int, default=[2039, 3457],
                        help='item_ids corresponding to metric_ids separated by spaces')
    parser.add_argument('-r', '--region_id', nargs='+', type=int, default=[100000100],
                        help='region id of the region')
    parser.add_argument('-s', '--source_ids', nargs='+', type=int, default=[35, 26],
                        help='specify the source_ids of the corresponding item-metric '
                             'separated by spaces')
    parser.add_argument('-f', '--frequency_ids', nargs='+', type=int, default=[1, 1],
                        help='frequency_ids of the corresponding gro-data_series separated by '
                             'spaces')
    parser.add_argument('-w', '--weights', nargs='+', type=float,
                        help='weights corresponding to the data_series separated by spaces')
    parser.add_argument('--initial_date', required=True, type=valid_date,
                        help='Format YYYY-MM-DD')
    parser.add_argument('--final_date', required=True, type=valid_date,
                        help='Format YYYY-MM-DD')
    parser.add_argument('--groapi_token', type=str, default=os.environ['GROAPI_TOKEN'],
                        help='GroAPI token')
    parser.add_argument('--api_host', type=str, default='api.gro-intelligence.com',
                        help='api host')
    parser.add_argument('--output_dir', type=str, default='')
    parser.add_argument('--report', type=str2bool, default=True,
                        help='Generates correlation matrix and scatter plots between ranks')
    parser.add_argument('--methods', nargs='+', type=str,
                        default=['euclidean', 'cumulative', 'ts-features'],
                        choices=['euclidean', 'cumulative', 'ts-features', 'dtw'],
                        help='methods of rank calculation. The arguments can be one or more of'
                             'the following strings - "euclidean", "ts features", "cumulative",'
                             '"dtw"')
    parser.add_argument('--start_date', type=valid_date, help='start date of all the Gro data '
                                                              'series to be used for this '
                                                              'analysis')
    parser.add_argument("--ENSO", action='store_true', help='Include ENSO for rank distance '
                                                            'calculation')
    parser.add_argument('--ENSO-weight', type=float, default=1, help='Weight of the ENSO index')
    parser.add_argument('--all_ranks', action='store_true', help='Lets you see all the ranks'
                                                                 'as opposed to separate method'
                                                                 'ranks')
    parser.add_argument('--num_jobs', type=int, default=0,
                        help='number of parallel processes for tsfresh')
    args = parser.parse_args()
    data_series_list = get_data_series_list(args.region_id, args.item_ids, args.metric_ids,
                                            args.source_ids, args.frequency_ids,
                                            api_token=args.groapi_token, api_host=args.api_host)
    folder_name = final_ranks_computation.get_file_name(args.groapi_token, data_series_list,
                                                        initial_date=args.initial_date,
                                                        final_date=args.final_date,
                                                        api_host=args.api_host)
    result = final_ranks_computation.analogous_years(
        args.groapi_token, data_series_list, args.initial_date, args.final_date,
        methods_list=args.methods, all_ranks=args.all_ranks, weights=args.weights, enso=args.ENSO,
        enso_weight=args.ENSO_weight, provided_start_date=args.start_date,
        tsfresh_num_jobs=args.num_jobs, api_host=args.api_host)
    final_ranks_computation.save_to_csv(args.groapi_token, result,
                                        folder_name,
                                        file_name='ranks.csv',
                                        output_dir=args.output_dir,
                                        api_host=args.api_host)
    if args.all_ranks and args.report:
        correlation_matrix = final_ranks_computation.generate_correlation_matrix(result)
        final_ranks_computation.save_to_csv(args.groapi_token, correlation_matrix, folder_name,
                                            file_name='correlation_matrix.csv',
                                            output_dir=args.output_dir, api_host=args.api_host)
        final_ranks_computation.generate_correlation_scatterplots(args.groapi_token, result,
                                                                  folder_name,
                                                                  output_dir=args.output_dir,
                                                                  api_host=args.api_host)
    return None


if __name__ == '__main__':
    main()



"""
This library contains utilities for
1. retrieving data series from gro api
2. extracting and transforming relevant part of the data series
"""

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
import pandas as pd


def get_data(client, metric_id, item_id, region_id, source_id, frequency_id, start_date):
    """
    Reads data from api into a dataframe then the values corresponding to non-unique dates are
    combined, resampled to daily frequency
    :param client: GroClient
    :param metric_id: Gro-metric
    :param item_id: Gro-item associated to the metric
    :param region_id: Gro-region
    :param source_id: Gro-source
    :param frequency_id: Gro-frequency
    :param start_date: start-date of the Gro data series user provided
    :return: A dataframe with an 'end_date' and 'value' column
    """
    data_series = {'metric_id': metric_id,
                   'item_id': item_id,
                   'region_id': region_id,
                   'source_id': source_id,
                   'frequency_id': frequency_id}
    client.add_single_data_series(data_series)
    data = client.get_df()[['end_date', 'value']]
    data = combine_subregions(data) # consolidates non unique dates together
    data = data.resample('D').nearest()
    data.loc[:, 'end_date'] = data.index
    start_date_bound = pd.to_datetime(start_date)
    data = data.loc[data.end_date >= start_date_bound][['end_date', 'value']]
    return data


def combine_subregions(df_sub_regions):
    """
    Consolidate values of sub_regions within the data frame
    by grouping together same date entries
    :param df_sub_regions: A dataframe with subregions
    :return: A dataframe with summed 'value' across regions for each date
    """
    if df_sub_regions['end_date'].is_unique:
        df_consolidated_regions = df_sub_regions
        df_consolidated_regions.index = df_sub_regions['end_date']
    else:
        df_consolidated_regions = df_sub_regions.groupby(['end_date'])[['value']].sum()
    df_consolidated_regions.index = pd.to_datetime(df_consolidated_regions.index)
    df_consolidated_regions.loc[:, 'end_date'] = df_consolidated_regions.index
    return df_consolidated_regions


def loop_initiation_dates(max_date, initial_date, final_date):
    """
    :param max_date: datetime
    :param initial_date: string 'YYYY-MM-DD'
    :param final_date: string 'YYYY-MM-DD'
    :return: dictionary with datetime objects as values
    """
    tz_info = max_date.tzinfo
    initial_date = parse(initial_date)
    initial_date = initial_date.replace(tzinfo=tz_info)
    final_date = parse(final_date)
    final_date = final_date.replace(tzinfo=tz_info)
    if initial_date >= final_date:
        raise ValueError('Initial date {} is not prior to final date {}. '
                         'Change and try again'.format(initial_date, final_date))
    if final_date >= initial_date + relativedelta(years=+1):
        raise ValueError('Initial date {} not within a year of final date {} '
                         'Change and try again'.format(initial_date, final_date))
    if final_date > max_date:
        raise ValueError('Data unavailable after {}'.format(max_date))
    loop_final_date = final_date
    loop_initial_date = initial_date
    while loop_final_date + relativedelta(years=+1) < max_date:
        loop_final_date = loop_final_date + relativedelta(years=+1)
        loop_initial_date = loop_initial_date + relativedelta(years=+1)
    return {'initial_date': loop_initial_date, 'final_date': loop_final_date}


def dates_to_period_string(date_1, date_2):
    """
    :param date_1: datetime object
    :param date_2: datetime object
    :return: string
    """
    return date_1.strftime("%Y-%m-%d") + ' to ' + date_2.strftime("%Y-%m-%d")


def extract_time_periods_by_dates(dataframe, initial_date, final_date):
    """
    Extract a dataframe from an input dataframe of daily values using the
    MM-DD for different years
    :param dataframe: A pandas dataframe
    :param final_date: 'YYYY-MM-DD'
    :param initial_date: 'YYYY-MM-DD'
    :return: A pandas dataframe
    """
    dataframe.loc[:, 'date'] = pd.to_datetime(dataframe['end_date'])
    max_date = dataframe['date'].max()
    min_date = dataframe['date'].min()
    loop_final_date = loop_initiation_dates(
        max_date, initial_date, final_date)['final_date']
    loop_initial_date = loop_initiation_dates(
        max_date, initial_date, final_date)['initial_date']
    extracted_df_list = []
    while loop_initial_date >= min_date:
        temp_df = dataframe[(dataframe['date'] >= loop_initial_date) &
                            (dataframe['date'] <= loop_final_date)]
        temp_df.loc[:, 'period'] = dates_to_period_string(loop_initial_date, loop_final_date)
        loop_initial_date = loop_initial_date + relativedelta(years=-1)
        loop_final_date = loop_final_date + relativedelta(years=-1)
        extracted_df_list.append(temp_df)
    extracted_df = pd.concat(extracted_df_list)
    extracted_df.loc[:, 'mm-dd'] = extracted_df['date'].dt.strftime("%m-%d")
    return extracted_df


def stack_time_periods_by_ddmm(dataframe):
    """
    A pivot table with 'period' as columns and 'mm-dd' as rows.
    :param dataframe: A pandas dataframe
    :return: A pandas dataframe
    """
    segmented_periods = pd.pivot_table(dataframe, values='value',
                                       index=['mm-dd'], columns=['period'])
    segmented_periods = segmented_periods[segmented_periods.index != '02-29']
    return segmented_periods



import os
from groclient import GroClient

def main():
    # set up the Batch Client, same as normal Client
    """ API Config """
    API_HOST = 'api.gro-intelligence.com'
    ACCESS_TOKEN = os.environ['GROAPI_TOKEN']

    api_client = GroClient(API_HOST, ACCESS_TOKEN)

    # specify everything except region_id
    selection = {
        'metric_id': 860032,
        'item_id': 274,
        'source_id': 25,
        'frequency_id': 9,
        'start_date': '1998-01-01T00:00:00.000Z',
        'end_date': '1998-01-01T00:00:00.000Z'
    }

    # make a list of this query for every county in Mississippi
    mississippi_county_ids = [region["id"] for region in api_client.get_descendant_regions(13075, 5)]

    selections = []
    for region_id in mississippi_county_ids:
        selection_temp = dict(selection)
        selection_temp["region_id"] = region_id
        selections.append(selection_temp)

    # make the request in a batch, asynchronous way
    output = api_client.batch_async_get_data_points(selections)

    # the output data is in the same order as the input queries. 
    for county_id, data in zip(mississippi_county_ids, output):
        if len(data) > 0: # some counties are missing data
            print("county_idx=%i produced %.0f tonnes of corn in 1998" % (county_id, data[0]["value"]))
        else:
            print("county_idx=%i has no data for 1998" % county_id)

if __name__ == "__main__":
        main()



"""Miscellaneous utility functions for operations not specific to the API client."""

try:
    # functools are native in Python 3.2.3+
    from functools import lru_cache as memoize
except ImportError:
    from backports.functools_lru_cache import lru_cache as memoize
import re
from math import ceil

from groclient.constants import DATA_SERIES_UNIQUE_TYPES_ID


@memoize(maxsize=None)
def str_camel_to_snake(term):
    """Convert a string from camelCase to snake_case.

    Parameters
    ----------
    term : string
        A camelCase string
    Returns
    -------
    string
        A new snake_case string

    """
    return re.sub(
        "([a-z0-9])([A-Z])", r"\1_\2", re.sub("(.)([A-Z][a-z]+)", r"\1_\2", term)
    ).lower()


@memoize(maxsize=None)
def str_snake_to_camel(term):
    """Convert a string from snake_case to camelCase.

    Parameters
    ----------
    term : string

    Returns
    -------
    string

    """
    camel = term.split("_")
    return "".join(camel[:1] + list([x[0].upper() + x[1:] for x in camel[1:]]))


def dict_reformat_keys(obj, format_func):
    """Convert a dictionary's keys from one string format to another.

    Parameters
    ----------
    obj : dict
        A dictionary with keys that need to be reformatted
    format_func : function
        Will execute on each key on the dict

    Returns
    -------
    dict
        A new dictionary with formatted keys

    """
    return {format_func(key): value for key, value in obj.items()}


def dict_unnest(obj):
    """Flatten a dictionary containing other dictionaries by concatenating their keys.

    Parameters
    ----------
    obj : dict
        A dictionary, which may or may not contain other dictionaries

    Returns
    -------
    dict
        A new dictionary, which has been reformatted

    """
    flat_obj = {}
    for key, value in obj.items():
        if isinstance(value, dict):
            for sub_key, sub_value in dict_unnest(value).items():
                flat_obj["{}_{}".format(key, sub_key)] = sub_value
        elif isinstance(value, list):
            for idx, sub_value in enumerate(value):
                flat_obj["{}_{}".format(key, idx)] = sub_value
        else:
            flat_obj[key] = value
    return flat_obj


def dict_assign(obj, key, value):
    """Chainable dictionary assignment. Returns a copy.

    Parameters
    ----------
    obj : dict
        A dictionary
    key : string
        Which attribute to set. May be a new attribute, or overwriting an existing one
    value
        A value of any type

    Returns
    -------
    dict
        A new dictionary, with the given attribute set to the given value

    """
    new_dict = dict(obj)
    new_dict[key] = value
    return new_dict


def list_chunk(arr, chunk_size=50):
    """Chunk an array into chunks of a given max length.

    Parameters
    ----------
    arr : list
    chunk_size : int, optional

    Returns
    -------
    list of lists

    """
    return [
        arr[i * chunk_size : (i + 1) * chunk_size]
        for i in range(int(ceil(len(arr) / float(chunk_size))))
    ]


def intersect(lhs_list, rhs_list):
    """Return the common elements of two lists

    Parameters
    ----------
    lhs_list : list
        A list of some type that can be compared using `in`
    rhs_list : list
        A list of some type that can be compared using `in`

    Returns
    -------
    list
        A list of common elements

    """
    return list(filter(lambda elem: elem in rhs_list, lhs_list))


def zip_selections(entity_ids, **optional_selections):
    """Take a list of entity_ids and optional named selections and create a "selections" dict

    Parameters
    ----------
    entity_ids : list of ints
        Must be in the order of DATA_SERIES_UNIQUE_TYPES_ID
    optional_selections : kwargs
        Additional options to be combined with entity selections. Like `insert_nulls`

    Returns
    -------
    dict

    """
    selection = dict(zip(DATA_SERIES_UNIQUE_TYPES_ID, entity_ids))

    # add optional pararms to selection
    for key, value in list(optional_selections.items()):
        if key not in DATA_SERIES_UNIQUE_TYPES_ID:
            selection[key] = value
    return selection



### These are the data series we should use when doing a similar regions similarity computation.

region_properties = {
    "soil_moisture": {
        "selected_entities": {
            'metric_id': 15531082,
            'item_id': 7382,
            'frequency_id': 1,
            'source_id': 43
        },
        "properties": {
            "type": "timeseries_fourier",
            "num_features": 150,
            "longest_period_feature_period": 365,
            "weight": 0.7,
            "weight_slope": True
        }
    },
    "land_surface_temperature": {
        "selected_entities": {
            'metric_id': 2540047,
            'item_id': 3457,
            'frequency_id': 1,
            'source_id': 26
        },
        "properties": {
            "type": "timeseries_fourier",
            "num_features": 150,
            "longest_period_feature_period": 365,
            "weight": 0.7,
            "weight_slope": False
        }
    },
    "rainfall": {
        "selected_entities": {
            'metric_id': 2100031,
            'item_id': 2039,
            'frequency_id': 1,
            'source_id': 35
        },
        "properties": {
            "type": "timeseries_fourier",
            "num_features": 150,
            "longest_period_feature_period": 365,
            "weight": 0.7,
            "weight_slope": True
        }
    },
    "cation_exchange_30cm": {
        "selected_entities": {
            'metric_id': 15531092, 
            'item_id': 9131, 
            'source_id': 89, 
            'frequency_id': 15
        },
        "properties": {
            "type": "pit", 
            "num_features": 1,
            "weight": 0.7
        }
    },
    "ph_h2o_30cm": {
        "selected_entities": {
            'metric_id': 15760040,
            'item_id': 9164,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    },
    "sand_30cm": {
        "selected_entities": {
            'metric_id': 15750042,
            'item_id': 9185,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    },
    "silt_30cm": {
        "selected_entities": {
            'metric_id': 15750042,
            'item_id': 9178,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    },
    "clay_30cm": {
        "selected_entities": {
            'metric_id': 15750042,
            'item_id': 9138,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    },
    "organic_carbon_content_fine_earth_30cm": {
        "selected_entities": {
            'metric_id': 15531050,
            'item_id': 9157,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    },
    "soil_water_capacity_100cm": {
        "selected_entities": {
            'metric_id': 15531082,
            'item_id': 9194,
            'source_id': 89,
            'frequency_id': 15
        },
        "properties": {
            "type": "pit",
            "num_features": 1,
            "weight": 0.7
        }
    }
}



from __future__ import division
from builtins import map
from builtins import str
from builtins import zip
import pandas
import unicodecsv
from api.client import crop_model

# Deprecated: moved to api.client.crop_model, use that directly.
class CropModel(crop_model.CropModel):

    def __init__(self, api_host, access_token):
        super(CropModel, self).__init__(api_host, access_token)
    



import platform
from pkg_resources import get_distribution

import mock
import numpy as np

from groclient import lib
from groclient.utils import dict_assign
from groclient.constants import DATA_SERIES_UNIQUE_TYPES_ID

MOCK_HOST = 'pytest.groclient.url'
MOCK_TOKEN = 'pytest.groclient.token'

LOOKUP_MAP = {
    'metrics': {
        '1': {'id': 1, 'name': 'metric 1', 'contains': [], 'belongsTo': [3], 'definition': 'def1'},
        '2': {'id': 2, 'name': 'metric 2', 'contains': [], 'belongsTo': [3], 'definition': 'def2'},
        '3': {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'definition': 'def3'},
        '4': {'id': 4, 'name': 'ancestor', 'contains': [3], 'belongsTo': [], 'definition': 'def4'}
    },
    'items': {
        '1': {'id': 1, 'name': 'item 1', 'contains': [], 'belongsTo': [3], 'definition': 'def1'},
        '2': {'id': 2, 'name': 'item 2', 'contains': [], 'belongsTo': [3], 'definition': 'def2'},
        '3': {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'definition': 'def3'},
        '4': {'id': 4, 'name': 'ancestor', 'contains': [3], 'belongsTo': [], 'definition': 'def4'}
    },
    'regions': {
        '1': {'id': 1, 'name': 'region 1', 'contains': [], 'belongsTo': [3], 'historical': True},
        '2': {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False},
        '3': {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'historical': False},
        '4': {'id': 4, 'name': 'ancestor', 'contains': [3], 'belongsTo': [], 'historical': False}
    },
    'frequencies': {},
    'sources': {},
    'units': {}
}


def initialize_requests_mocker_and_get_mock_data(mock_requests_get, mock_data={
    'data': [
        {'name': 'obj1'},
        {'name': 'obj2'},
        {'name': 'obj3'}
    ]}
):
    mock_requests_get.return_value.json.return_value = mock_data
    mock_requests_get.return_value.status_code = 200
    return mock_data


@mock.patch('requests.get')
def test_get_available(mock_requests_get):
    mock_data = initialize_requests_mocker_and_get_mock_data(mock_requests_get)

    # Test data
    entity_types = ['items', 'metrics', 'regions']

    for ent_type in entity_types:
        assert lib.get_available(MOCK_TOKEN, MOCK_HOST, ent_type) == mock_data['data']


@mock.patch('requests.get')
def test_list_available(mock_requests_get):
    # Tests the base functionality
    mock_data = initialize_requests_mocker_and_get_mock_data(mock_requests_get)

    entities = {'metricId': '123', 'itemId': '456', 'regionId': '789'}

    assert lib.list_available(MOCK_TOKEN, MOCK_HOST, entities) == mock_data['data']


@mock.patch('requests.get')
def test_list_available_snake_to_camel(mock_requests_get):
    # Tests that the camel-ing fix is working properly.
    mock_data = initialize_requests_mocker_and_get_mock_data(mock_requests_get)
    entities = {'metric_id': '123', 'item_id': '456', 'regionId': '789'}
    assert lib.list_available(MOCK_TOKEN, MOCK_HOST, entities) == mock_data['data']


@mock.patch('requests.get')
def test_single_lookup(mock_requests_get):
    api_response = {
        'data': {
            '12345': {
                'id': 12345,
                'name': 'Vegetables',
                'contains': [67890],
                'belongsTo': []
            }
        }
    }
    initialize_requests_mocker_and_get_mock_data(mock_requests_get, api_response)
    expected_return = {'id': 12345, 'name': 'Vegetables', 'contains': [67890], 'belongsTo': []}
    assert lib.lookup(MOCK_TOKEN, MOCK_HOST, 'items', 12345) == expected_return


@mock.patch('requests.get')
def test_multiple_lookups(mock_requests_get):
    api_response = {
        'data': {
            '12345': {'id': 12345, 'name': 'Vegetables', 'contains': [67890], 'belongsTo': []},
            '67890': {'id': 67890, 'name': 'Eggplant', 'contains': [], 'belongsTo': [12345]}
        }
    }
    initialize_requests_mocker_and_get_mock_data(mock_requests_get, api_response)
    expected_return = {
        '12345': {'id': 12345, 'name': 'Vegetables', 'contains': [67890], 'belongsTo': []},
        '67890': {'id': 67890, 'name': 'Eggplant', 'contains': [], 'belongsTo': [12345]}
    }
    assert lib.lookup(MOCK_TOKEN, MOCK_HOST, 'items', [12345, 67890]) == expected_return


@mock.patch('requests.get')
def test_lookup_with_numpy(mock_requests_get):
    api_response = {
        'data': {
            '12345': {'id': 12345, 'name': 'Vegetables', 'contains': [67890], 'belongsTo': []},
            '67890': {'id': 67890, 'name': 'Eggplant', 'contains': [], 'belongsTo': [12345]}
        }
    }
    initialize_requests_mocker_and_get_mock_data(mock_requests_get, api_response)
    expected_return = {
        '12345': {'id': 12345, 'name': 'Vegetables', 'contains': [67890], 'belongsTo': []},
        '67890': {'id': 67890, 'name': 'Eggplant', 'contains': [], 'belongsTo': [12345]}
    }
    assert lib.lookup(MOCK_TOKEN, MOCK_HOST, 'items', np.array([12345, 67890])) == expected_return

    assert lib.lookup(MOCK_TOKEN, MOCK_HOST, 'items',
                      np.array([12345])[0]) == expected_return['12345']


@mock.patch('requests.get')
def test_get_data_series(mock_requests_get):
    # Test general case
    mock_data = initialize_requests_mocker_and_get_mock_data(mock_requests_get)
    selection_dict = {'item_id': 123, 'metric_id': 456, 'region_id': 789,
                      'partner_region_id': 161718, 'frequency_id': 101112, 'source_id': 12}

    assert lib.get_data_series(MOCK_TOKEN, MOCK_HOST, **selection_dict) == mock_data['data']


@mock.patch('requests.get')
def test_get_data_points(mock_requests_get):
    list_of_series_format_data = [{
        'series': {},
        'data': [['2000-01-01', '2000-12-31', 1]]
    }]
    single_series_format_data = [{
        'start_date': '2000-01-01',
        'end_date': '2000-12-31',
        'value': 1,
        'unit_id': None,
        'input_unit_id': None,
        'input_unit_scale': 1,
        'reporting_date': None,
        'metric_id': None,
        'item_id': None,
        'metadata': {},
        'region_id': None,
        'partner_region_id': 0,
        'frequency_id': None,
        # 'source_id': None, TODO: add source to output
    }]
    initialize_requests_mocker_and_get_mock_data(mock_requests_get,
                                                 mock_data=list_of_series_format_data)

    # Test data
    selection_dict = {'item_id': 123, 'metric_id': 456, 'region_id': 789,
                      'frequency_id': 101112, 'source_id': 131415, 'partner_region_id': 161718}

    assert lib.get_data_points(MOCK_TOKEN, MOCK_HOST, **selection_dict) == single_series_format_data


def test_list_of_series_to_single_series():
    assert lib.list_of_series_to_single_series([{
        'series': {
            'metricId': 1,
            'itemId': 2,
            'regionId': 3,
            'unitId': 4,
            'inputUnitId': 5,
            'belongsTo': {'itemId': 22},
            'metadata': {'includesHistoricalRegion': True}
        },
        'data': [
            ['2001-01-01', '2001-12-31', 123],
            ['2002-01-01', '2002-12-31', 123, '2012-01-01'],
            ['2003-01-01', '2003-12-31', 123, None, 15, {'confInterval': 2}]
        ]
    }], add_belongs_to=True) == [
        {'start_date': '2001-01-01',
         'end_date': '2001-12-31',
         'value': 123,
         'unit_id': 4,
         'metadata': {},
         'input_unit_id': 4,
         'input_unit_scale': 1,
         'reporting_date': None,
         'metric_id': 1,
         'item_id': 2,
         'region_id': 3,
         'partner_region_id': 0,
         'frequency_id': None,
         'belongs_to': {'item_id': 22}},
        {'start_date': '2002-01-01',
         'end_date': '2002-12-31',
         'value': 123,
         'unit_id': 4,
         'metadata': {},
         'input_unit_id': 4,
         'input_unit_scale': 1,
         'reporting_date': '2012-01-01',
         'metric_id': 1,
         'item_id': 2,
         'region_id': 3,
         'partner_region_id': 0,
         'frequency_id': None,
         'belongs_to': {'item_id': 22}},
        {'start_date': '2003-01-01',
         'end_date': '2003-12-31',
         'value': 123,
         'unit_id': 15,
         'metadata': {'conf_interval': 2},
         'input_unit_id': 15,
         'input_unit_scale': 1,
         'reporting_date': None,
         'metric_id': 1,
         'item_id': 2,
         'region_id': 3,
         'partner_region_id': 0,
         'frequency_id': None,
         'belongs_to': {'item_id': 22}},
    ]

    # test the add_belongs_to kwarg:
    assert 'belongs_to' in lib.list_of_series_to_single_series([{
        'series': {'unitId': 4, 'belongsTo': {'itemId': 22}},
        'data': [['2001-01-01', '2001-12-31', 123]]
    }], add_belongs_to=True)[0]

    assert 'belongs_to' not in lib.list_of_series_to_single_series([{
        'series': {'unitId': 4, 'belongsTo': {'itemId': 22}},
        'data': [['2001-01-01', '2001-12-31', 123]]
    }], add_belongs_to=False)[0]

    # test the include_historical kwarg:
    assert len(lib.list_of_series_to_single_series([{
        'series': {'unitId': 4, 'belongsTo': {'itemId': 22}, 'metadata': {'includesHistoricalRegion': True}},
        'data': [['2001-01-01', '2001-12-31', 123]]
    }], include_historical=False)) == 0

    assert len(lib.list_of_series_to_single_series([{
        'series': {'unitId': 4, 'belongsTo': {'itemId': 22}, 'metadata': {'includesHistoricalRegion': True}},
        'data': [['2001-01-01', '2001-12-31', 123]]
    }], include_historical=True)) == 1

    # test invalid input propagation
    assert lib.list_of_series_to_single_series('test input') == 'test input'


@mock.patch('requests.get')
def test_search(mock_requests_get):
    mock_data = ['obj1', 'obj2', 'obj3']
    mock_data = initialize_requests_mocker_and_get_mock_data(mock_requests_get, mock_data=mock_data)

    assert lib.search(MOCK_TOKEN, MOCK_HOST, 'items', 'test123') == mock_data


@mock.patch('groclient.lib.lookup')
@mock.patch('groclient.lib.search')
def test_search_and_lookup(search_mocked, lookup_mocked):
    # Set up
    # mock return values
    search_mocked.return_value = [{'id': 1}, {'id': 2}]
    lookup_mocked.side_effect = lookup_mock
    search_and_lookup_result = list(lib.search_and_lookup(MOCK_TOKEN, MOCK_HOST,
                                                          'regions', 'test123'))

    assert search_and_lookup_result == [
        LOOKUP_MAP['regions']['1'],
        LOOKUP_MAP['regions']['2']
    ]


@mock.patch('groclient.lib.lookup')
@mock.patch('requests.get')
def test_lookup_belongs(mock_requests_get, lookup_mocked):
    mock_requests_get.return_value.json.return_value = {'data': {'1': [3]}}
    mock_requests_get.return_value.status_code = 200
    lookup_mocked.side_effect = lookup_mock

    lookup_belongs_result = list(lib.lookup_belongs(MOCK_TOKEN, MOCK_HOST, 'regions', 1))

    assert lookup_belongs_result == [LOOKUP_MAP['regions']['3']]


@mock.patch('requests.get')
def test_get_source_ranking(mock_requests_get):
    mock_return = [60, 14, 2, 1]
    mock_requests_get.return_value.json.return_value = mock_return
    mock_requests_get.return_value.status_code = 200

    query_parameters = {'item_id': 1, 'metric_id': 2, 'region_id': 3, 'frequency_id': 4}

    ranked_sources_list = lib.get_source_ranking(MOCK_TOKEN, MOCK_HOST, query_parameters)
    assert len(ranked_sources_list) == 4


@mock.patch('requests.get')
def test_rank_series_by_source(mock_requests_get):
    # for each series selection, mock ranking of 3 source ids
    mock_return = [11, 123, 33]
    mock_requests_get.return_value.json.return_value = mock_return
    mock_requests_get.return_value.status_code = 200

    full_data_series = {
        'metric_id': 2540047,
        'item_id': 3457,
        'region_id': 13474,
        'partner_region_id': 56789,
        'source_id': 123,
        'source_name': 'dontcare',
        'frequency_id': 9,
        'start_date': '2020-01-01',
        'end_date': '2020-05-01',
        'metadata': {'historical': True}
    }
    # input selection should allow a list of ids
    partial_selection = {
        'metric_id': 2540047,
        'item_id': 1457,
        'region_id': [13474,13475]
    }

    expected = [
        full_data_series,
        dict_assign(partial_selection, 'source_id', mock_return[0]),
        dict_assign(partial_selection, 'source_id', mock_return[1]),
        dict_assign(partial_selection, 'source_id', mock_return[2])
    ]
    for idx, series in enumerate(lib.rank_series_by_source(MOCK_TOKEN,
                                                           MOCK_HOST,
                                                           [full_data_series,
                                                            partial_selection])):
        assert series == expected[idx]


def lookup_mock(MOCK_TOKEN, MOCK_HOST, entity_type, entity_ids):
    if isinstance(entity_ids, int):
        return LOOKUP_MAP[entity_type][str(entity_ids)]
    if isinstance(entity_ids, list):
        return {str(entity_id): LOOKUP_MAP[entity_type][str(entity_id)]
                for entity_id in entity_ids}


@mock.patch('groclient.lib.lookup')
@mock.patch('requests.get')
def test_get_descendant(mock_requests_get, lookup_mocked):
    mock_requests_get.return_value.json.return_value = {'data': {'4': [1, 2, 3]}}
    mock_requests_get.return_value.status_code = 200
    lookup_mocked.side_effect = lookup_mock

    assert lib.get_descendant(MOCK_TOKEN, MOCK_HOST, 'items', 4) == [
        {'id': 1, 'name': 'item 1', 'contains': [], 'belongsTo': [3], 'definition': 'def1'},
        {'id': 2, 'name': 'item 2', 'contains': [], 'belongsTo': [3], 'definition': 'def2'},
        {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'definition': 'def3'}
    ]
    
    assert lib.get_descendant(MOCK_TOKEN, MOCK_HOST, 'metrics', 4, include_details=True) == [
        {'id': 1, 'name': 'metric 1', 'contains': [], 'belongsTo': [3], 'definition': 'def1'},
        {'id': 2, 'name': 'metric 2', 'contains': [], 'belongsTo': [3], 'definition': 'def2'},
        {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'definition': 'def3'}
    ]

    assert lib.get_descendant(MOCK_TOKEN, MOCK_HOST, 'items', 4, include_details=False) == [
        {'id': 1}, {'id': 2}, {'id': 3}
    ]

    mock_requests_get.return_value.json.return_value = {'data': {'4': [3]}}
    assert lib.get_descendant(MOCK_TOKEN, MOCK_HOST, 'items', 4, distance=1) == [
        {'id': 3, 'name': 'parent', 'contains': [1, 2], 'belongsTo': [4], 'definition': 'def3'}
    ]

    

@mock.patch('groclient.lib.lookup')
@mock.patch('requests.get')
def test_descendant_regions(mock_requests_get, lookup_mocked):
    mock_requests_get.return_value.json.return_value = {'data': {'3': [1, 2]}}
    mock_requests_get.return_value.status_code = 200
    lookup_mocked.side_effect = lookup_mock

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3) == [
        {'id': 1, 'name': 'region 1', 'contains': [], 'belongsTo': [3], 'historical': True},
        {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3, include_details=True) == [
        {'id': 1, 'name': 'region 1', 'contains': [], 'belongsTo': [3], 'historical': True},
        {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3, include_details=False) == [
        {'id': 1}, {'id': 2}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3, include_historical=True) == [
        {'id': 1, 'name': 'region 1', 'contains': [], 'belongsTo': [3], 'historical': True},
        {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3, include_historical=False) == [
        {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3,
                                      include_historical=True, include_details=True) == [
        {'id': 1, 'name': 'region 1', 'contains': [], 'belongsTo': [3], 'historical': True},
        {'id': 2, 'name': 'region 2', 'contains': [], 'belongsTo': [3], 'historical': False}
    ]

    assert lib.get_descendant_regions(MOCK_TOKEN, MOCK_HOST, 3, include_historical=False,
                                      include_details=False) == [{'id': 2}]


@mock.patch('requests.get')
def test_get_top(mock_requests_get):
    mock_response = [
        {'itemId': 274, 'value': 13175206696, 'unitId': 14},
        {'itemId': 574, 'value': 13175206878, 'unitId': 14},
        {'itemId': 7193, 'value': 13175206343, 'unitId': 14}
    ]
    mock_requests_get.return_value.json.return_value = mock_response
    mock_requests_get.return_value.status_code = 200
    assert lib.get_top(MOCK_TOKEN, MOCK_HOST, 'items', metric_id=14) == mock_response
    assert lib.get_top(MOCK_TOKEN, MOCK_HOST, 'items', num_results=3, metric_id=14) == mock_response


@mock.patch('requests.get')
def test_get_geo_centre(mock_requests_get):
    US_data = {
        "regionId": 1215,
        "regionName": "United States",
        "centre": [
            39.8333,
            -98.5855
        ]
    }
    api_response = {
        'data': [US_data]
    }
    initialize_requests_mocker_and_get_mock_data(mock_requests_get, api_response)
    assert lib.get_geo_centre(MOCK_TOKEN, MOCK_HOST, 1215) == [US_data]


@mock.patch('requests.get')
def test_get_geo_jsons(mock_requests_get):
    api_response = {
        'data': [{
            "regionId": 13051,
            "regionName": "Alabama",
            "centre": [
                32.7933,
                -86.8278
            ],
            "geojson": {
                "type": "MultiPolygon",
                "coordinates": [[[[ -88.201896667, 35.0088806150001]]]]
            }
        },
        {
            "regionId": 13052,
            "regionName": "Alaska",
            "centre": [
                64.2386,
                -152.279
            ],
            "geojson": {
                "type": "MultiPolygon",
                "coordinates": [[[[ -179.07043457, 51.2564086920001]]]]
            }
        }]
    }
    expected_return = [{
        "region_id": 13051,
        "region_name": "Alabama",
        "centre": [
            32.7933,
            -86.8278
        ],
        "geojson": {
            "type": "MultiPolygon",
            "coordinates": [[[[ -88.201896667, 35.0088806150001]]]]
        }
    },
    {
        "region_id": 13052,
        "region_name": "Alaska",
        "centre": [
            64.2386,
            -152.279
        ],
        "geojson": {
            "type": "MultiPolygon",
            "coordinates": [[[[ -179.07043457, 51.2564086920001]]]]
        }
    }]
    initialize_requests_mocker_and_get_mock_data(mock_requests_get, api_response)
    assert lib.get_geojsons(MOCK_TOKEN, MOCK_HOST, 1215, None, 7) == expected_return


@mock.patch('groclient.lib.get_geojsons')
def test_get_geo_json(geojsons_mocked):
    geojsons_mocked.return_value = [{
        "region_id": 1215,
        "region_name": "United States",
        "centre": [
            39.8333,
            -98.5855
        ],
        "geojson": "{\"type\":\"GeometryCollection\",\"geometries\":[{\"type\":\"MultiPolygon\",\"coordinates\":[[[[-155.651382446,20.1647224430001]]]]}]}"
    }]
    expected_return = {
        'type': 'GeometryCollection',
        'geometries': [{'type': 'MultiPolygon', 'coordinates': [[[[-155.651382446, 20.1647224430001]]]]}]
    }
    assert lib.get_geojson(MOCK_TOKEN, MOCK_HOST, 1215, 7) == expected_return



import numpy as np
from numpy.testing import assert_almost_equal
import pandas as pd
from pandas.util.testing import assert_frame_equal
import pytest

from api.client.samples.analogous_years.lib import feature_extractions


def create_test_data():
    test_df = pd.DataFrame({'rows': ['row_1', 'row_2', 'row_3'],
                            'col_1': [1, 2, 3],
                            'col_2': [4, 4, 4],
                            'col_3': [7, 8, np.nan]}).set_index('rows')
    return test_df


def test_rm_const_cols():
    expected_df = create_test_data().drop('col_2', axis=1)
    expected_df = expected_df.drop('col_3', axis=1)
    assert_frame_equal(feature_extractions.rm_const_cols(create_test_data()), expected_df)


def test_feature_scaling():
    expected_df = create_test_data()
    for column in list(create_test_data().columns):
        expected_df.loc[:, column] = expected_df[column] - (create_test_data()[column].mean())
        if create_test_data()[column].to_numpy().std() != 0:
            std = np.nanstd(create_test_data()[column].to_numpy())
            expected_df.loc[:, column] = expected_df[column] / std
        else:
            expected_df.loc[:, column] = 0.0
    expected_df = feature_extractions.rm_const_cols(expected_df)
    assert_almost_equal(feature_extractions.feature_scaling(create_test_data()),
                        np.array(expected_df))


def test_cumulative():
    valid_data = create_test_data()
    valid_data.index = ['row_1', 'row_1', 'row_2']
    valid_data.loc[:, 'period'] = valid_data.index
    valid_data.loc[:, 'value'] = valid_data['col_1']
    expected = pd.DataFrame({'period': ['row_1', 'row_2'],
                             'value': [3, 3]}).set_index('period')
    assert_frame_equal(feature_extractions.cumulative(valid_data, 'period', 'value'), expected)


def test_pca_transformation():
    null_data = create_test_data()
    with pytest.raises(ValueError):
        feature_extractions.pca_transformation(null_data)



from __future__ import division
from builtins import map
from builtins import zip
from datetime import datetime
import numpy
import pandas
from groclient.client import GroClient


class CropModel(GroClient):

    def compute_weights(self, crop_name, metric_name, regions):
        """Compute a vector of 'weights' that can be used for crop-weighted
        average across regions, as in :meth:`~.compute_crop_weighted_series`.

        For each region, the weight of is the mean value over time, of
        the given metric for the given crop, normalized so the sum
        across all regions is 1.0.

        For example: say we have a `region_list = [{'id': 1, 'name':
        'Province1'}, {'id': 2, 'name': 'Province2'}]`. This could
        be a list returned by :meth:`~.client.search_and_lookup` or
        :meth:`~.client.get_descendant_regions` for example.  Now say
        `model.compute_weights('soybeans', 'land cover area',
        region_list)` returns `[0.6, 0.4]`, that means Province1
        has 60% and province2 has 40% of the total area planted across
        the two regions, when averaged across all time.

        Parameters
        ----------
        crop_name : string
        metric_name : string
        regions : list of dicts
            Each entry is a region with id and name

        Returns
        -------
        list of floats
           weights corresponding to the regions.


        See also
        --------
        :meth:`~.compute_crop_weighted_series`

        """
        # Get the weighting series
        entities = { 'item_ids': set(), 'metric_ids': set() }
        for region in regions:
            series = self.add_data_series(
                item=crop_name, metric=metric_name, region=region['name'],
                result_filter=lambda x: 'region_id' not in x or \
                                        x['region_id'] == region['id'])
            if series:
                entities['item_ids'].add(series['item_id'])
                entities['metric_ids'].add(series['metric_id'])

        # Compute the average over time for reach region
        df = self.get_df()

        def mapper(region):
            return df[(df['item_id'].isin(entities['item_ids'])) &
                      (df['metric_id'].isin(entities['metric_ids'])) &
                      (df['region_id'] == region['id'])]['value'].mean(skipna=True)
        means = list(map(mapper, regions))
        self.get_logger().debug('Means = {}'.format(
            list(zip([region['name'] for region in regions], means))))
        # Normalize into weights
        total = numpy.nansum(means)
        if not numpy.isclose(total, 0.0):
            return [float(mean)/total for mean in means]
        self.get_logger().warning(
            'Cannot normalize {} {} data.'.format(crop_name, metric_name))
        return means


    def compute_crop_weighted_series(self, weighting_crop_name, weighting_metric_name,
                                     item_name, metric_name, regions,
                                     weighting_func=lambda w, v: w*v):
        """Compute the 'crop-weighted average' of the series for the given
        item and metric, across regions. The weight of a region is the
        fraction of the value of the weighting series represented by
        that region as explained in :meth:`~.compute_weights`.

        For example: say we have a `region_list = [{'id': 1, 'name':
        'Province1'}, {'id': 2, 'name': 'Province2'}]`. This could
        be a list returned by :meth:`~.client.search_and_lookup` or
        client.get_descendant_regions for example.  Now
        `model.compute_crop_weighted_series('soybeans', 'land cover
        area', 'vegetation ndvi', 'vegetation indices index',
        region_list)` will return a dataframe where the NDVI of each
        province is multiplied by the fraction of total soybeans
        area is accounted for by that province. Thus taking the sum
        across provinces will give a crop weighted average of NDVI.

        Parameters
        ----------
        weighting_crop_name : string
        weighting_metric_name : string
        item_name : string
        metric_name : string
        regions : list of dicts
            Each entry is a region with id and name
        weighting_func: optional function
            A function of (weight, value) to apply. Default: weight*value

        Returns
        -------
        pandas.DataFrame
            contains the data series for the given item_name, metric_name,
            for each region in regions, with values adjusted
            by the crop weight for that region.

        """
        weights = self.compute_weights(
            weighting_crop_name, weighting_metric_name, regions)

        entities = { 'item_ids': set(), 'metric_ids': set() }
        for region in regions:
            series = self.add_data_series(
                item=item_name, metric=metric_name, region=region['name'],
                result_filter=lambda x: 'region_id' not in x or \
                                        x['region_id'] == region['id'])
            if series:
                entities['item_ids'].add(series['item_id'])
                entities['metric_ids'].add(series['metric_id'])

        df = self.get_df()
        series_list = []
        for (region, weight) in zip(regions, weights):
            self._logger.info(u'Computing {}_{}_{} x {}'.format(
                item_name, metric_name,  region['name'], weight))
            series = df[(df['item_id'].isin(entities['item_ids'])) &
                        (df['metric_id'].isin(entities['metric_ids'])) &
                        (df['region_id'] == region['id'])].copy()
            series.loc[:, 'value'] = weighting_func(weight, series['value'])
            # TODO: change metric to reflect it is weighted in this copy
            series_list.append(series)
        return pandas.concat(series_list)

    def compute_gdd(self, tmin_series, tmax_series, base_temperature,
                    start_date, end_date, min_temporal_coverage,
                    upper_temperature_cap):
        """Compute Growing Degree Days value from specific data series.

        This function performs the low-level computation used in :meth:`~.growing_degree_days`.

        Parameters
        ----------
        tmin_series : dict
            A data series object for min temperature
            e.g. {metric_id: 1, item_id: 2, region_id: 3, source_id: 4, frequency_id: 5}
        tmax_series : dict
            A data series object for max temperature
            e.g. {metric_id: 1, item_id: 2, region_id: 3, source_id: 4, frequency_id: 5}
        base_temperature : number
        start_date : string
            YYYY-MM-DD date
        end_date : string
            YYYY-MM-DD date
        min_temporal_coverage : float, optional
        upper_temperature_cap : number, optional

        Returns
        -------
        number
            The sum of the GDD over all days in the interval

        See also
        --------
        :meth:`~.growing_degree_days`
        """
        self.add_single_data_series(tmin_series)
        self.add_single_data_series(tmax_series)
        df = self.get_df()
        if df is None or df.empty:
            raise Exception("Insufficient data for GDD")
        # For each day we want (t_min + t_max)/2, or more generally,
        # the average temperature for that day.
        tmean = df.loc[(df.item_id == tmax_series['item_id']) | \
                       (df.item_id == tmin_series['item_id'])].groupby(
                           ['region_id', 'metric_id', 'frequency_id',
                            'start_date', 'end_date']).mean()
        duration = datetime.strptime(end_date, '%Y-%m-%d') - \
                   datetime.strptime(start_date, '%Y-%m-%d')
        if duration.days > 366:
            self.get_logger().warning(
                'GDD time range is more than 1 year {} - {}.'.format(
                    start_date, end_date))
        coverage_threshold = min_temporal_coverage * duration.days
        if tmean.value.size < coverage_threshold:
            raise Exception(
                "Insufficient coverage for GDD, {} < {} data points. ".format(
                    tmean.value.size, coverage_threshold) +
                "min_temporal_coverage is {}.".format(min_temporal_coverage))
        gdd_values = tmean.value.apply(
            lambda x: max(min(x, upper_temperature_cap) - base_temperature, 0))
        # TODO: group by freq and normalize in case not daily
        # TODO: add unit conversions in case future sources are in different units
        return gdd_values.sum()

    def growing_degree_days(self, region_name, base_temperature,
                            start_date, end_date, min_temporal_coverage=1.0,
                            upper_temperature_cap=float("Infinity")):
        """Get Growing Degree Days (GDD) for a region.

        Growing degree days (GDD) are a weather-based indicator that
        allows for assessing crop phenology and crop development,
        based on heat accumulation. GDD for one day is defined as
        max(T_mean - T_base, 0), where T_mean is the average
        temperature of that day if available. Typically T_mean is
        approximated as (T_max + T_min)/2. If upper_temperature_cap is
        specified, T_mean is capped to not exceed that value.

        The GDD over a longer time interval is the sum of the GDD over
        all days in the interval. Days where the data is missing
        contribute 0 GDDs, i.e. are treated as if T_mean = T_base.
        Use the temporal coverage threshold to avoid computing GDD
        with too little data.

        The threshold and the base temperature should be carefuly
        selected based on fundamental understanding of the crops and
        region of interest.

        The region can be any region of the Gro regions, from a point
        location to a district, province etc. This will use the best
        available data series for T_max and T_min for the given region
        and time period, using "find_data_series". In the simplest
        case, if the given region is a weather station location which
        has data for the time period, then that will be used. If it's
        a district or other region, the underlying data could be from
        one or more weather stations and/or satellite.  To by-pass the
        search for available series, use :meth:`~.compute_gdd` directly.

        Parameters
        ----------
        region_name : string
        base_temperature : number
        start_date : string
            YYYY-MM-DD date
        end_date : string
            YYYY-MM-DD date
        min_temporal_coverage : float, optional
        upper_temperature_cap : number, optional

        Returns
        -------
        number
            The sum of the GDD over all days in the interval

        """
        try:
            tmin_series = next(self.find_data_series(
                item='Temperature min', metric='Temperature', region=region_name,
                start_date=start_date, end_date=end_date))
            tmax_series = next(self.find_data_series(
                item='Temperature max', metric='Temperature', region=region_name,
                start_date=start_date, end_date=end_date))
            return self.compute_gdd(tmin_series, tmax_series, base_temperature,
                                    start_date, end_date, min_temporal_coverage,
                                    upper_temperature_cap)
        except StopIteration:
            raise Exception(
                "Can't find data series to compute GDD in region {}".format(
                    region_name))
